# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
import uuid
from enum import Enum
from future.builtins import str, list, int
import logging

from astrophysix.simdm.utils import ObjectList, GalacticaValidityCheckMixin, DataType
from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable
from astrophysix import units as U


log = logging.getLogger("astrophysix.simdm")


class PropertyFilterFlag(Enum):
    NO_FILTER = ("no_filter", "Not used in filters")
    BASIC_FILTER = ("basic_filter", "Filter in basic form")
    ADVANCED_FILTER = ("advanced_filter", "Filter in advanced form")

    def __init__(self, filter_flag, display_name):
        self._flag = filter_flag
        self._disp_name = display_name

    @property
    def flag(self):
        """Object property filter flag value"""
        return self._flag

    @property
    def displayed_flag(self):
        """Object property filter flag displayed name"""
        return self._disp_name

    @classmethod
    def from_flag(cls, flag):
        """
        Parameters
        ----------
        flag: :obj:`string`
            property filter flag value

        Returns
        -------
        t: :class:`~astrophysix.simdm.catalogs.targobj.PropertyFilterFlag`
            Property filter flag matching the requested flag value.

        Raises
        ------
        ValueError
            if requested flag value does not match any property filter flag.

        Example
        -------
            >>> flag = PropertyFilterFlag.from_flag("no_filter")
            >>> flag.displayed_flag
            "Not used in filters"
            >>> flag2 = PropertyFilterFlag.from_flag("MY_UNKNOWN_FLAG")
            ValuerError: No PropertyFilterFlag defined with the flag 'MY_UNKNOWN_FLAG'.
        """
        for fflag in cls:
            if fflag.flag == flag:
                return fflag
        raise ValueError("No PropertyFilterFlag defined with the flag '{f:s}'".format(f=flag))


class PropertySortFlag(Enum):
    NO_SORT = ("no_sort", "Not used for sorting")
    BASIC_SORT = ("basic_sort", "Sort in basic form")
    ADVANCED_SORT = ("advanced_sort", "Sort in advanced form")

    def __init__(self, sort_flag, display_name):
        self._flag = sort_flag
        self._disp_name = display_name

    @property
    def flag(self):
        """Object property sort flag value"""
        return self._flag

    @property
    def displayed_flag(self):
        """Object property sort flag displayed name"""
        return self._disp_name

    @classmethod
    def from_flag(cls, flag):
        """
        Parameters
        ----------
        flag: :obj:`string`
            property sort flag value

        Returns
        -------
        t: :class:`~astrophysix.simdm.catalogs.targobj.PropertySortFlag`
            Property sort flag matching the requested flag value.

        Raises
        ------
        ValueError
            if requested flag value does not match any property sort flag.

        Example
        -------
            >>> flag = PropertySortFlag.from_flag("basic_sort")
            >>> flag.displayed_flag
            "Sort in basic form"
            >>> flag2 = PropertySortFlag.from_flag("MY_UNKNOWN_FLAG")
            ValuerError: No PropertySortFlag defined with the flag 'MY_UNKNOWN_FLAG'.
        """
        for sflag in cls:
            if sflag.flag == flag:
                return sflag
        raise ValueError("No PropertySortFlag defined with the flag '{f:s}'".format(f=flag))


class ObjectProperty(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Target object property class (Simulation data model)

    Parameters
    ----------
    property_name: :obj:`string`
        property name (mandatory)
    description: :obj:`string`
        object property description
    unit: :obj:`string` or :class:`~astrophysix.units.unit.Unit`
        object property physical unit
    filter_flag: :class:`~astrophysix.simdm.catalogs.targobj.PropertyFilterFlag`
        target object property filter flag. Default :attr:`PropertyFilterFlag.NO_FILTER <astrophysix.simdm.catalogs.targobj.PropertyFilterFlag.NO_FILTER>`
    sort_flag: :class:`~astrophysix.simdm.catalogs.targobj.PropertySortFlag`
        target object property sort flag. Default :attr:`PropertySortFlag.NO_SORT <astrophysix.simdm.catalogs.targobj.PropertySortFlag.NO_SORT>`
    dtype: :class`simdm.utils.DataType`
        property data type. Default DataType.REAL.
    """
    def __init__(self, **kwargs):
        super(ObjectProperty, self).__init__(**kwargs)
        self._prop_name = ""
        self._description = ""

        # Target object property name
        if "property_name" not in kwargs:
            raise AttributeError("{cname:s} 'property_name' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.property_name = kwargs["property_name"]

        # Target object property description
        if "description" in kwargs:
            self.description = kwargs["description"]

        # Target object property filter/sort flags
        self._filter_flag = PropertyFilterFlag.NO_FILTER
        if "filter_flag" in kwargs:
            self.filter_flag = kwargs["filter_flag"]
        self._sort_flag = PropertySortFlag.NO_SORT
        if "sort_flag" in kwargs:
            self.sort_flag = kwargs["sort_flag"]

        # Target object property datatype
        self._dtype = DataType.REAL
        if "dtype" in kwargs:
            self.datatype = kwargs["dtype"]

        # Target object property unit
        self._unit = U.none
        if "unit" in kwargs:
            self.unit = kwargs["unit"]

    def __eq__(self, other):
        """
        ObjectProperty comparison method

        other: :class:`~astrophysix.simdm.catalogs.targobj.ObjectProperty`
            target object property instance to compare to
        """
        if not super(ObjectProperty, self).__eq__(other):
            return False

        # Compare property name
        if self._prop_name != other.property_name:
            return False

        # Compare property description
        if self._description != other.description:
            return False

        # Compare datatype
        if self._dtype != other.datatype:
            return False

        # Compare sort/filter flags
        if self._filter_flag != other.filter_flag or self._sort_flag != other.sort_flag:
            return False

        # Compare units
        return self._unit == other.unit and self._unit.name == other.unit.name

    @property
    def property_name(self):
        """Target object property name. Can be edited."""
        return self._prop_name

    @property_name.setter
    def property_name(self, new_prop_name):
        try:
            self._prop_name = Stringifiable.cast_string(new_prop_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'property_name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def display_name(self):
        """Object property display name. Concatenation of the property name and its unit LaTex formula, if defined."""
        if self._unit == U.none:
            return self._prop_name
        d = "{n:s} ({ul:s})".format(n=self._prop_name, ul=self._unit.latex)
        return d

    @property
    def datatype(self):
        """Object property datatype (:class:`~astrophysix.simdm.utils.DataType`)"""
        return self._dtype

    @datatype.setter
    def datatype(self, dtype):
        try:
            tk = Stringifiable.cast_string(dtype)
            self._dtype = DataType.from_key(tk)
        except ValueError as ve:
            err_msg = "Object property 'datatype' error : {verr:s}".format(verr=str(ve))
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:  # Not a valid string
            if not isinstance(dtype, DataType):
                err_msg = "Object property 'datatype' attribute is not a valid DataType enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._dtype = dtype

    @property
    def description(self):
        """Object property description. Can be edited."""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def filter_flag(self):
        """
        Object property filter flag. Can be edited.

        Returns
        -------
        f: :class:`~atrophysix.simdm.catalogs.targob.PropertyFilterFlag`
            object property filter flag
        """
        return self._filter_flag

    @filter_flag.setter
    def filter_flag(self, new_fflag):
        try:
            fflag = Stringifiable.cast_string(new_fflag)
            self._filter_flag = PropertyFilterFlag.from_flag(fflag)
        except ValueError as ve:
            err_msg = "Object property 'filter_flag' property error : {verr:s}.".format(verr=str(ve))
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:  # Not a valid string
            if not isinstance(new_fflag, PropertyFilterFlag):
                err_msg = "Object property 'filter_flag' attribute is not a valid PropertyFilterFlag enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._filter_flag = new_fflag

    @property
    def sort_flag(self):
        """
        Object property sort flag. Can be edited.

        Returns
        -------
        f: :class:`~atrophysix.simdm.catalogs.targob.PropertySortFlag`
            object property sort flag
        """
        return self._sort_flag

    @sort_flag.setter
    def sort_flag(self, new_sflag):
        try:
            sflag = Stringifiable.cast_string(new_sflag)
            self._sort_flag = PropertySortFlag.from_flag(sflag)
        except ValueError as ve:
            err_msg = "Object property 'sort_flag' property error : {verr:s}.".format(verr=str(ve))
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:  # Not a valid string
            if not isinstance(new_sflag, PropertySortFlag):
                err_msg = "Object property 'sort_flag' attribute is not a valid PropertySortFlag enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._sort_flag = new_sflag

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, new_unit):
        if isinstance(new_unit, U.Unit):
            self._unit = new_unit
        else:
            try:
                s = Stringifiable.cast_string(new_unit, valid_empty=False)
                self._unit = U.Unit.from_name(s)
            except TypeError:  # Not a valid string
                err_msg = "Object property 'unit' property is not a valid (non-empty) string."
                log.error(err_msg)
                raise AttributeError(err_msg)
            except AttributeError as aerr:
                err_msg = "Object property 'unit' property error : {uerr:s}.".format(uerr=str(aerr))
                log.error(err_msg)
                raise AttributeError(err_msg)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["property_name", "description", "filter_flag", "sort_flag", "dtype", "unit"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a ObjectProperty object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the ObjectProperty into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(ObjectProperty, self)._hsp_write(h5group, **kwargs)

        # Write object property name
        self._hsp_write_attribute(h5group, ('name', self._prop_name), **kwargs)

        # Write target object property description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write object property filter/sort flags
        self._hsp_write_attribute(h5group, ('filter_flag', self._filter_flag.flag), **kwargs)
        self._hsp_write_attribute(h5group, ('sort_flag', self._sort_flag.flag), **kwargs)

        # Write property type
        self._hsp_write_attribute(h5group, ("datatype", self._dtype.key), **kwargs)

        # Write object property unit
        self._hsp_write_attribute(h5group, ('unit', self._unit), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a ObjectProperty instance from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        prop: ``ObjectProperty``
            Read ObjectProperty instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(ObjectProperty, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read catalog name
        name = cls._hsp_read_attribute(h5group, 'name', "generic result name")

        # Read property filter/sort flags
        fflag = cls._hsp_read_attribute(h5group, "filter_flag", "object property filter flag")
        sflag = cls._hsp_read_attribute(h5group, "sort_flag", "object property sort flag")

        dtype = cls._hsp_read_attribute(h5group, "datatype", "object property type", raise_error_if_not_found=False)
        if dtype is None:
            dtype = DataType.REAL

        # Read parameter setting unit
        u = cls._hsp_read_unit(h5group, "unit")

        # Create target object property
        prop = cls(uid=uid, property_name=name, unit=u, filter_flag=fflag, sort_flag=sflag, dtype=dtype)

        # Read target object property description, if defined
        pdesc = cls._hsp_read_attribute(h5group, 'description', "target object property description",
                                        raise_error_if_not_found=False)
        if pdesc is not None:
            prop.description = pdesc

        return prop

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check property name length
        if len(self._prop_name) > 64:
            log.warning("{t!s} name is too long for Galactica (max. 64 characters).".format(t=self))

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{pname:s}' target object property".format(pname=self._prop_name)


class ObjectPropertyGroup(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Target object property group class (Simulation data model)

    Parameters
    ----------
    group_name: property group name (mandatory)
    description: property group description
    """
    def __init__(self, **kwargs):
        super(ObjectPropertyGroup, self).__init__(**kwargs)
        self._group_name = ""
        self._description = ""
        self._group_properties = ObjectList(ObjectProperty, 'property_name')

        # Target object property group name
        if "group_name" not in kwargs:
            raise AttributeError("{cname:s} 'group_name' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.group_name = kwargs["group_name"]

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        ObjectPropertyGroup comparison method

        other: :class:`~astrophysix.simdm.catalogs.targobj.ObjectPropertyGroup`
            target object property group instance to compare to
        """
        if not super(ObjectPropertyGroup, self).__eq__(other):
            return False

        if self._group_name != other.group_name:
            return False

        if self._description != other.description:
            return False

        if self._group_properties != other.group_properties:
            return False

        return True

    @property
    def group_name(self):
        """Object property group name. Can be edited."""
        return self._group_name

    @group_name.setter
    def group_name(self, new_grp_name):
        try:
            self._group_name = Stringifiable.cast_string(new_grp_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'group_name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def group_properties(self):
        """Object property group :class:`~astrophysix.simdm.catalogs.targobj.ObjectProperty` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._group_properties

    @property
    def description(self):
        """Object property group description. Can be edited."""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["group_name", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a ObjectPropertyGroup object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the ObjectPropertyGroup into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(ObjectPropertyGroup, self)._hsp_write(h5group, **kwargs)

        # Write object property group name
        self._hsp_write_attribute(h5group, ('name', self._group_name), **kwargs)

        # Write target object property group description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write group properties UUID list
        self._hsp_write_object_list(h5group, "GROUP_PROPERTIES", self._group_properties, "group_prop_",
                                    uid_list_only=True, **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a ObjectProperty instance from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        prop: ``ObjectProperty``
            Read ObjectProperty instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(ObjectPropertyGroup, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read object property group name
        name = cls._hsp_read_attribute(h5group, 'name', "generic result name")

        # Create target object property group
        pgroup = cls(uid=uid, group_name=name)

        # ---------------------- Read target object property group members (object properties) ----------------------- #
        # Search for already instantiated target object properties in dependency object dictionary
        if dependency_objdict is None:
            err_msg = "Cannot find any target object property already instantiated in the TargetObject instance."
            log.error(err_msg)
            raise IOError(err_msg)

        # Get dictionary of target object properties of the corresponding class :
        if ObjectProperty.__name__ not in dependency_objdict:
            err_msg = "Cannot find any {cname:s} instance.".format(cname=ObjectProperty.__name__)
            log.error(err_msg)
            raise IOError(err_msg)
        obj_prop_dict = dependency_objdict[ObjectProperty.__name__]

        # Find object property
        for spuid in ObjectProperty._hsp_read_object_list(h5group, "GROUP_PROPERTIES", "group_prop_", "group property",
                                                          uid_list_only=True, dependency_objdict=dependency_objdict):
            objprop_uid = uuid.UUID(spuid)

            # Find target object property according to its UUID
            if objprop_uid not in obj_prop_dict:
                err_msg = "Cannot find {cname:s} instance with uid {uid:s}.".format(cname=ObjectProperty.__name__,
                                                                                    uid=spuid)
                log.error(err_msg)
                raise IOError(err_msg)

            p = obj_prop_dict[objprop_uid]
            pgroup.group_properties.add(p)
        # ------------------------------------------------------------------------------------------------------------ #

        # Read target object property group description, if defined
        pdesc = cls._hsp_read_attribute(h5group, 'description', "target object property group description",
                                        raise_error_if_not_found=False)
        if pdesc is not None:
            pgroup.description = pdesc

        return pgroup

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check property group name length
        if len(self._group_name) > 32:
            log.warning("{t!s} name is too long for Galactica (max. 32 characters).".format(t=self))

        # Pointless to run the Galactica validity check on the group properties ? If it is already done at the
        # TargetObject level...

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{gname:s}' property group".format(gname=self._group_name)


class TargetObject(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Catalog target object class (Simulation data model)

    Parameters
    ----------
    name: :obj:`string`
        object name (mandatory)
    description: :obj:`string`
        result description
    """
    def __init__(self, **kwargs):
        super(TargetObject, self).__init__(**kwargs)
        self._name = ""
        self._description = ""
        self._obj_properties = ObjectList(ObjectProperty, "property_name")
        self._obj_properties.add_deletion_handler(self._can_delete_object_property)
        self._obj_property_groups = ObjectList(ObjectPropertyGroup, "group_name",
                                               object_addition_vcheck=(self._can_add_objet_property_in_a_group,
                                                                       "group_properties"))
        self._obj_property_groups.add_validity_check_method(self._can_add_prop_group)

        # Target object name
        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        TargetObject comparison method

        other: :class:`~astrophysix.simdm.catalogs.targobj.TargetObject`
            target object instance to compare to
        """
        if not super(TargetObject, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._description != other.description:
            return False

        # Compare opbject properties and property groups
        if self._obj_properties != other.object_properties:
            return False
        if self._obj_property_groups != other.property_groups:
            return False

        return True

    @property
    def name(self):
        """Target object name. Can be edited."""
        return self._name

    @name.setter
    def name(self, new_cat_name):
        try:
            self._name = Stringifiable.cast_string(new_cat_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def object_properties(self):
        """Target object :class:`~astrophysix.simdm.catalogs.targobj.ObjectProperty` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._obj_properties

    @property
    def property_groups(self):
        """Target object :class:`~astrophysix.simdm.catalogs.targobj.ObjectPropertyGroup` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._obj_property_groups

    @property
    def description(self):
        """Target object description. Can be edited."""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_add_prop_group(self, prop_group):
        """
        Checks that a given ObjectPropertyGroup instance can be added into this Target object group list.
        Verifies that each one of the ObjectProperty instance of this group already belongs to the TargetObject object
        property list. Raises an AttributeError if not.

        Parameters
        ----------
        prop_group: ``astrophysix.simdm.catalogs.targobj.ObjectPropertyGroup``
            Object property group to add
        """
        for obj_prop in prop_group.group_properties:
            self._can_add_objet_property_in_a_group(obj_prop)

    def _can_add_objet_property_in_a_group(self, obj_prop):
        """
        Checks that a given ObjectProperty instance can be added into any of this TargetObject's ObjectPropertyGroup.
        Verifies that the ObjectProperty instance belongs to the TargetObject object propert list. Raises an
        AttributeError if not.

        Parameters
        ----------
        obj_prop: ``astrophysix.simdm.catalogs.targobj.ObjectProperty``
            Object property to add
        """
        if obj_prop not in self._obj_properties:
            err_msg = "{op!s} does not belong to this {cname:s} object property " \
                      "list.".format(cname=self.__class__.__name__, op=obj_prop)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_delete_object_property(self, obj_prop):
        """
        Checks if an object property is not linked to any object property group and can be safely deleted.
        Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        obj_prop: ``:class:~astrophysix.simdm.catalogs.targobj.ObjectProperty``
            target object property about to be deleted

        Returns
        -------
        o: str or None
        """
        prop_list = []
        for pgroup in self._obj_property_groups:
            if obj_prop in pgroup.group_properties:
                prop_list.append("{s!s} - {pg!s} - {op!s}".format(s=self, pg=pgroup, op=obj_prop))

        if len(prop_list) > 0:
            return prop_list

        return None

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a TargetObject object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the TargetObject into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(TargetObject, self)._hsp_write(h5group, **kwargs)

        # Write target object name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write target object description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write target object properties
        self._hsp_write_object_list(h5group, "OBJ_PROPERTIES", self._obj_properties, "obj_prop_", **kwargs)

        # Write target object property groups
        self._hsp_write_object_list(h5group, "OBJ_PROPERTY_GROUPS", self._obj_property_groups, "obj_prop_group_",
                                    **kwargs)

        self._hsp_write_callback(str(self), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a TargetObject instance from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        targobj: ``TargetObject``
            Read TargetObject instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(TargetObject, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read target object name
        name = cls._hsp_read_attribute(h5group, 'name', "generic result name")

        # Create target object
        targobj = cls(uid=uid, name=name)

        # Build object property list and add each property into its target object property list + build target object
        # property dictionary indexed by their UUID
        if ObjectProperty.__name__ not in dependency_objdict:
            dependency_objdict[ObjectProperty.__name__] = {}
        obj_prop_dict = dependency_objdict[ObjectProperty.__name__]
        for p in ObjectProperty._hsp_read_object_list(h5group, "OBJ_PROPERTIES", "obj_prop_", "target object property",
                                                      dependency_objdict=dependency_objdict):
            obj_prop_dict[p.uid] = p
            targobj.object_properties.add(p)

        # Build object property group list and add each property group into the target object
        for pgroup in ObjectPropertyGroup._hsp_read_object_list(h5group, "OBJ_PROPERTY_GROUPS", "obj_prop_group_",
                                                                "target object property groups",
                                                                dependency_objdict=dependency_objdict):
            targobj.property_groups.add(pgroup)

        # Read target object description, if defined
        targobj_desc = cls._hsp_read_attribute(h5group, 'description', "target object description",
                                               raise_error_if_not_found=False)
        if targobj_desc is not None:
            targobj.description = targobj_desc

        return targobj

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check target object name
        if len(self._name) > 32:
            log.warning("{t!s} name is too long for Galactica (max. 32 characters).".format(t=self))

        # Check target object property and object property groups validity
        self._obj_properties.galactica_validity_check(**kwargs)
        self._obj_property_groups.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{targobj_name:s}' target object".format(targobj_name=self._name)


__all__ = ["ObjectProperty", "ObjectPropertyGroup", "TargetObject", "PropertySortFlag", "PropertyFilterFlag"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.simdm.datafiles.file.AssociatedFile

.. autoclass:: astrophysix.simdm.datafiles.file.FitsFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass:: astrophysix.simdm.datafiles.file.PickleFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass::  astrophysix.simdm.datafiles.file.AsciiFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass:: astrophysix.simdm.datafiles.file.HDF5File
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass::  astrophysix.simdm.datafiles.file.JsonFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass::  astrophysix.simdm.datafiles.file.CSVFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass::  astrophysix.simdm.datafiles.file.YamlFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass::  astrophysix.simdm.datafiles.file.TarGzFile
    :members: filename, last_modified, load_file, save_to_disk, raw_file_data
    :show-inheritance: AssociatedFile

"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, int
import logging
import os
import datetime
import numpy as N

from astrophysix.utils import FileType, FileUtil
from astrophysix.utils.dt_utils import DatetimeUtil
from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable
from ..utils import GalacticaValidityCheckMixin


log = logging.getLogger("astrophysix.simdm")


class AssociatedFile(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    FILE_MD5SUM_HDF5_ATTR_NAME = "file_md5sum"
    FILE_RAW_DATA_ATTR_NAME = "file_raw_data"
    FILE_TYPE = FileType.ASCII_FILE

    def __init__(self, **kwargs):
        """
        Datafile associated file base class (Simulation data model)

        Parameters
        ----------
        filename: :obj:`string`
            file name (mandatory)
        file_md5sum: :obj:`string`
            MD5SUM of the file (mandatory)
        """
        super(AssociatedFile, self).__init__(**kwargs)

        self._filename = ""
        self._file_md5sum = ""
        self._last_modif_time = None
        self._raw_file_data = None

        # ------------------------------------------------ File name ------------------------------------------------- #
        if "filename" not in kwargs:
            err_msg = "{cname:s} 'filename' attribute is not defined (mandatory).".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        fname = kwargs["filename"]
        if self.FILE_TYPE != FileUtil.get_file_type(fname):  # File type does not match defined file type
            err_msg = "'{fname!s}' is not a valid {fta:s} file name".format(fname=fname, fta=self.FILE_TYPE.alias)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._filename = fname
        # ------------------------------------------------------------------------------------------------------------ #

        # ------------------------------------------------ File md5sum ----------------------------------------------- #
        if "file_md5sum" not in kwargs:
            err_msg = "{cname:s} 'file_md5sum' attribute is not defined " \
                      "(mandatory).".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        try:
            self._file_md5sum = Stringifiable.cast_string(kwargs["file_md5sum"])
        except TypeError:
            err_msg = "{cname:s} 'file_md5sum' attribute is not a valid string".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        # ------------------------------------------------------------------------------------------------------------ #

        # -------------------------------------- File last modification datetime ------------------------------------- #
        if "modif_dt" not in kwargs:
            err_msg = "{cname:s} 'modif_dt' attribute is not defined (mandatory).".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        dt = kwargs["modif_dt"]
        if not isinstance(dt, datetime.datetime):
            err_msg = "{cname:s} 'modif_dt' attribute is not a valid datetime instance."
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._last_modif_time = dt
        # ------------------------------------------------------------------------------------------------------------ #

        # ------------------------------------------ Raw file data bytes --------------------------------------------- #
        if "raw_file_data" in kwargs:
            rfd = kwargs["raw_file_data"]
            if not isinstance(rfd, N.void):
                err_msg = "{cname:s} 'raw_file_data' attribute is not a valid numpy void object."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._raw_file_data = rfd
        # ------------------------------------------------------------------------------------------------------------ #

    @property
    def filename(self):
        """Gets associated file name. Cannot be edited."""
        return self._filename

    @property
    def file_md5sum(self):
        """Datafile associated file MD5sum getter."""
        return self._file_md5sum

    @property
    def last_modified(self):
        """Returns file last modification time. Cannot be edited."""
        return self._last_modif_time

    @property
    def raw_file_data(self):
        """
        File binary raw data. Cannot be edited.

        Returns
        -------
        raw_data: :obj:`bytes`
        """
        # Force file data loading if not already read from HDF5 file
        self._hsp_lazy_read()

        return self._raw_file_data.tobytes()

    def __eq__(self, other):
        """
        Rich comparison method for associated files

        Parameters
        ----------
        other: ``AssociatedFile``
            associated file to compare to:
        """
        # Avoid Hdf5StudyPersistent comparison test => it compares the UUID !
        # if not super(AssociatedFile, self).__eq__(other):
        #     return False
        # Compares only the equality of the object classes
        if self.__class__ != other.__class__:
            return False

        if self._filename != other.filename:
            return False

        if self._file_md5sum != other.file_md5sum:
            return False

        if self._last_modif_time != other.last_modified:
            return False

        return True

    @classmethod
    def _all_subclasses(cls):
        """
        Returns a list of all sub(sub*)classes of a given class (recursive class method)

        Returns
        -------
        subclass_list: ``list``
            list of all subclasses of a given class
        """
        subclass_list = cls.__subclasses__()
        other_subclasses = [c for subclass in subclass_list for c in subclass._all_subclasses()]
        return other_subclasses + subclass_list

    @classmethod
    def load_file(cls, filepath):
        """
        Loads an :class:`~astrophysix.simdm.datafiles.file.AssociatedFile` object from a filepath

        Parameters
        ----------
        filepath: :obj:`string`
            path of the file to load.

        Returns
        -------
        f: :class:`~astrophysix.simdm.datafiles.file.AssociatedFile` instance
            Loaded associatedfile
        """
        file_path = Stringifiable.cast_string(filepath, valid_empty=False)
        fpath = FileUtil.locate_file(file_path)
        filename = os.path.basename(fpath)
        ftype = FileUtil.get_file_type(fpath)
        if ftype != cls.FILE_TYPE:
            err_msg = "Invalid filename for a {ft:s} file ({fp!s}).".format(ft=cls.FILE_TYPE.alias, fp=file_path)
            log.error(err_msg)
            raise AttributeError(err_msg)
        file_md5sum = FileUtil.md5sum(fpath)
        last_modified = DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath))
        with open(file_path, 'rb') as fp:
            raw_file_data = N.void(fp.read())

        f = cls(filename=filename, file_md5sum=file_md5sum, modif_dt=last_modified, raw_file_data=raw_file_data)

        return f

    def _export_file(self, filepath):
        with open(filepath, 'wb') as fp:
            fp.write(self.raw_file_data)

    def save_to_disk(self, filepath=None):
        """
        Save associated file to an external file on the local filesystem

        Parameters
        ----------
        filepath: :obj:`string`
            external file path
        """
        if filepath is not None:
            file_path = Stringifiable.cast_string(filepath, valid_empty=False)
            fpath = FileUtil.new_filepath(file_path, self.FILE_TYPE, create_dir=True)
            ftype = FileUtil.get_file_type(filepath)
            if ftype != self.FILE_TYPE:
                err_msg = "Invalid filename for a {ft:s} file ({fp!s}).".format(ft=self.FILE_TYPE.alias, fp=filepath)
                log.error(err_msg)
                raise AttributeError(err_msg)
        else:
            fpath = os.path.abspath(self._filename)

        # Force file data loading if not already read from HDF5 file
        self._hsp_lazy_read()

        # Do the actual file saving operation...
        self._export_file(fpath)
        log.info("File '{fp:s}' saved.".format(fp=fpath))

        # Update file 'Last modified' time and 'Last access' time
        lm_ts = DatetimeUtil.utc_to_timestamp(self._last_modif_time)
        os.utime(fpath, (-1, lm_ts))  # Set last access time to 'now' and last modified time to original modif. time.

        # Validity check (Md5 sum comparison)
        if self._file_md5sum != FileUtil.md5sum(fpath):
            err_msg = "MD5 sum validity check failed upon {ft:s} file export into " \
                      "'{fp!s}'.".format(ft=self.FILE_TYPE.alias, fp=fpath)
            log.error(err_msg)
            raise IOError(err_msg)

        return fpath

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["filename", "file_md5sum", "modif_dt", "raw_file_data"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a AssociatedFile object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the AssociatedFile into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(AssociatedFile, self)._hsp_write(h5group, **kwargs)

        # Write AssociatedFile attributes
        self._hsp_write_attribute(h5group, ("filename", self._filename), **kwargs)
        modif_tms = DatetimeUtil.utc_to_timestamp(self._last_modif_time)
        self._hsp_write_attribute(h5group, ("last_modif_date", modif_tms), **kwargs)

        # Write file raw data
        raw_data_attribute_name = "_raw_file_data"  # => self._raw_file_data
        write_dataset = self._hsp_write_dataset(h5group, (self.FILE_RAW_DATA_ATTR_NAME, raw_data_attribute_name),
                                                md5sum_params=(self.FILE_MD5SUM_HDF5_ATTR_NAME, self._file_md5sum),
                                                **kwargs)
        if write_dataset:
            self._hsp_write_callback(str(self), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a AssociatedFile object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        af: ``AssociatedFile``
            Read AssociatedFile instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(AssociatedFile, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read AssociatedFile info from HDF5
        # ftype = FileType.from_alias(cls._hsp_read_attribute(h5group, "file_type", "associated file type"))
        fname = cls._hsp_read_attribute(h5group, "filename", "associated file name")
        file_md5sum = cls._hsp_read_attribute(h5group, cls.FILE_MD5SUM_HDF5_ATTR_NAME, "associated file MD5SUM")
        mod_dt = cls._hsp_read_attribute(h5group, "last_modif_date", "associated file last modification datetime")
        mod_dt_utc = DatetimeUtil.utc_from_timestamp(mod_dt)
        # af = cls(uid=uid, file_type=ftype, filename=fname, file_md5sum=file_md5sum, modif_dt=mod_dt)
        af = cls(uid=uid, filename=fname, file_md5sum=file_md5sum, modif_dt=mod_dt_utc)

        # Set HDF5 group/file info for lazy I/O
        af._hsp_set_lazy_source(h5group)

        return af

    def _hsp_lazy_read_data(self, h5group):
        """
        Lazy read method to load RawBinaryFile from HDF5 file (*.h5)

        Parameters
        ----------
        h5group: `h5py.Group`
        """
        log.info("Reading {ft:s} data from {gname:s} ({f!s} file).".format(ft=self.FILE_TYPE.alias, gname=h5group.name,
                                                                           f=h5group.file.filename))

        # Read raw data file, if any + md5sum
        self._raw_file_data = self._hsp_read_dataset(h5group, self.FILE_RAW_DATA_ATTR_NAME,
                                                     "{ft:s} file raw data".format(ft=self.FILE_TYPE.alias),
                                                     raise_error_if_not_found=True)

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        # Galactica validity check for associated file
        pass

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = "[{fname:s}] datafile associated {ft:s} file".format(fname=self._filename, ft=self.FILE_TYPE.alias)
        return s


class FitsFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.FITS_FILE` file class.
    """
    FILE_TYPE = FileType.FITS_FILE


class PickleFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.PICKLE_FILE` file class.
    """
    FILE_TYPE = FileType.PICKLE_FILE


class TarGzFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.TARGZ_FILE` file class.
    """
    FILE_TYPE = FileType.TARGZ_FILE


class JsonFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.JSON_FILE` file class.
    """
    FILE_TYPE = FileType.JSON_FILE


class CSVFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.CSV_FILE` file class.
    """
    FILE_TYPE = FileType.CSV_FILE


class AsciiFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.ASCII_FILE` file class.
    """
    FILE_TYPE = FileType.ASCII_FILE


class YamlFile(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.YAML_FILE` file class.
    """
    FILE_TYPE = FileType.YAML_FILE


class HDF5File(AssociatedFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.HDF5_FILE` file class.
    """
    FILE_TYPE = FileType.HDF5_FILE


__all__ = ["AssociatedFile", "PickleFile", "TarGzFile", "JsonFile", "CSVFile", "AsciiFile", "HDF5File", "YamlFile"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import unicode_literals, print_function
from future.builtins import str
import pytest
from astrophysix import units as U


class TestUnit(object):
    def test_express_exceptions(self):
        """
        Test exceptions raised by Unit.express(method)
        """
        # Check incompatible types
        with pytest.raises(U.UnitError) as e:
            f = U.Msun.express(U.kpc/U.Myr)
        assert str(e.value) == "Incompatible dimensions between :\n- Msun : (1.9889e+30 kg) (type: mass) and\n" \
                               "- (977792 m.s^-1) (type: velocity)"

        # Check Unit instance
        with pytest.raises(AttributeError) as e:
            f = U.Myr.express(3.9)
        assert str(e.value) == "'unit' attribute must be a 'Unit' instance."

    def test_express_valid_value(self):
        """
        Test conversion factor returned by Unit.express() method
        """
        f = U.kpc.express(U.ly)
        assert f == 3261.56377714188

    def test_unit_multiplication_error(self):
        """
        Test Unit instance multiplication raises UnitError
        """
        with pytest.raises(U.UnitError) as e:
            a = U.hour * ("Apple", 1, "Dog")
        assert str(
            e.value) == "Unable to multiply a Unit instance by something which is neither a Unit object nor a " \
                        "scalar."

    def test_unit_multiplication(self):
        """
        Test Unit instance multiplication
        """
        # Check left/right multiplication yields identical results
        a = U.c * U.Myr
        s = "{unit!s}".format(unit=a)  # call to __str__() here
        assert s == "(9.46073e+21 m)"
        assert a == U.Myr * U.c

        # Check real value multiplication
        a = 1000.0 * U.pc
        assert a == U.kpc

        # Check coefficient approximation
        a = 1000.0011 * U.pc  # More than 1.0E-6 relative tolerance on equality !
        assert a != U.kpc

    def test_unit_equality_identity(self):
        """
        Tests Unit instance equality and (strict) identity
        """
        assert U.km_s == U.km/U.s
        assert not U.km_s.identical(U.km/U.s)
        u = U.Unit.create_unit(name="Msun_pc2", base_unit=U.Msun/U.pc**2, descr="Solar mass per square parsec",
                               latex="\\textrm{M}_{\\odot}\\cdot\\textrm{pc}^{-2}")
        assert not u.identical(U.Msun/U.pc**2)
        assert U.Unit.from_name("Msun_pc2") == U.Msun/U.pc**2
        assert u.description == "Solar mass per square parsec"

    def test_unit_division(self):
        """
        Test Unit instance division
        """
        v = U.km / U.s
        s = "{unit!s}".format(unit=v)  # call to __str__() here
        assert s == "(1000 m.s^-1)"
        assert v != U.s / U.km  # Left/right division should not yield identical results !

    def test_unit_division_error(self):
        """
        Test Unit instance division raises UnitError
        """
        with pytest.raises(U.UnitError) as e:
            a = U.hour / ("Apple", 1, "Dog")
        assert str(e.value) == "Unable to divide a Unit instance by something which is neither a Unit object nor a " \
                               "scalar."

    def test_unit_power(self):
        """
        Test Unit instance to a given exponent
        """
        a = (10.0 * U.km) ** 3
        s = "{unit!s}".format(unit=a)  # call to __str__() here
        assert s == "(1e+12 m^3)"

    def test_base_SI_unit(self):
        """
        Test Unit.is_base_unit() method
        """
        assert U.kg.is_base_unit()
        assert not U.kpc.is_base_unit()
        assert U.m.is_base_unit()
        assert not U.Msun.is_base_unit()

    def test_physical_type(self):
        """
        Test Unit instance physical types
        """
        assert U.km.physical_type == "length"
        assert U.mGauss.physical_type == "magnetic flux density"
        assert U.Msun.physical_type == "mass"

        a = U.kg / U.km ** 3
        assert a.physical_type == "volume density"

        H = U.T * U.m ** 2 / U.A  # Henry
        assert H.physical_type == "inductance"

        a = U.mGauss / U.V / U.kg ** 2  # Really exotic physical quantity
        assert a.physical_type == U.Unit.UNKNOWN_PHYSICAL_TYPE

    def test_unit_instance_pickup_from_registry(self):
        """
        Test Unit instance retrieval from base Unit registry
        """
        assert U.Unit.from_name("Msun") is U.Msun

    def test_unit_iterator(self):
        """
        Test Unit registry iteration method
        """
        assert len(list(U.Unit.iterate_units())) == 91  # 91 units defined in Unit registry
        assert len(list(U.Unit.iterate_units(phys_type=U.kg.physical_type))) == 6  # 6 mass units defined in Unit registry
        assert U.kpc in U.Unit.iterate_units(phys_type=U.m.physical_type)
        assert U.Msun not in U.Unit.iterate_units(phys_type=U.W.physical_type)


__all__ = ["TestUnit"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str
import pytest
import uuid
import os
import datetime
import sys
from astrophysix.simdm import SimulationStudy, Project, ProjectCategory


class TestStudy(object):
    def test_01_study_init(self):
        """
        Tests Study instance initialisation
        """
        # Tests study defined without a valid Project instance raises an exception
        with pytest.raises(AttributeError, match="SimulationStudy 'project' attribute is not a valid Project object."):
            study_exc = SimulationStudy(project="Hi, My name is Robert !")  # Not a Project instance

        # Study initialisation => Ok should not raise  an error
        proj = Project(category=ProjectCategory.Cosmology, project_title="My awesome project", alias="NEW_PROJ")
        study = SimulationStudy(project=proj)

        assert isinstance(study.uid, uuid.UUID)
        assert isinstance(study.creation_time, datetime.datetime)
        assert isinstance(study.last_modification_time, datetime.datetime)
        assert study.project is proj

    def test_02_study_hdf5_io(self, tmp_path):
        """
        Tests saving/loading SimulationStudy from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        proj = Project(category=ProjectCategory.Cosmology, project_title="My awesome project", alias="NEW_PROJ")
        study = SimulationStudy(project=proj)

        fname = str(tmp_path / "study.h5")
        study.save_HDF5(study_fname=fname)
        study_loaded = SimulationStudy.load_HDF5(fname)

        print("study UUID : {uid!s}".format(uid=study.uid))
        assert study.uid == study_loaded.uid

        print("Study creation time : {t:s}".format(t=study.creation_time.isoformat()))
        assert study.creation_time == study_loaded.creation_time

        print("Study last modification time : {t:s}".format(t=study.last_modification_time.isoformat()))
        assert study.last_modification_time == study_loaded.last_modification_time

        assert study.project == study_loaded.project

    def test_03_study_hdf5_cross_io(self):
        """
        Tests cross-Python version loading SimulationStudy from HDF5 file (files saved by test_02_study_hdf5_io() method)
        """
        if sys.version_info.major > 2:  # Python 3.+
            # Load from a Python 2-saved HDF5 study
            study_loaded = SimulationStudy.load_HDF5(os.path.join(os.path.dirname(__file__), "io", "python_2", "study.h5"))
            assert study_loaded.creation_time.isoformat() == "2019-11-12T12:35:59.239199+00:00"
            assert study_loaded.last_modification_time.isoformat() == "2019-11-12T12:35:59.247400+00:00"
            assert str(study_loaded.uid) == "690a292b-8c49-4e97-9bd5-71a2159c3924"

            proj = Project(uid=uuid.UUID("df50c695-ad50-4c31-8e9d-1b1327645c22"), category=ProjectCategory.Cosmology,
                           project_title="My awesome project", alias="NEW_PROJ")
        else:  # Python 2.x
            # Load from a Python 3-saved HDF5 study
            study_loaded = SimulationStudy.load_HDF5(os.path.join(os.path.dirname(__file__), "io", "python_3", "study.h5"))
            assert study_loaded.creation_time.isoformat() == "2019-11-12T12:34:35.958546+00:00"
            assert study_loaded.last_modification_time.isoformat() == "2019-11-12T12:34:35.963422+00:00"
            assert str(study_loaded.uid) == "9a3549b2-711a-4f6d-8ece-8cdb203b8dff"

            proj = Project(uid=uuid.UUID("e82c5602-8272-4e6e-b85a-272096980d65"), category=ProjectCategory.Cosmology,
                           project_title="My awesome project", alias="NEW_PROJ")

        study = SimulationStudy(project=proj)
        assert study.project == study_loaded.project


__all__ = ["TestStudy"]




# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.simdm.datafiles.plot.PlotType
    :members:
    :undoc-members:

.. autoclass:: astrophysix.simdm.datafiles.plot.PlotInfo
    :members:

"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import json
import os
import tempfile
import webbrowser

from future.builtins import str, list, int
import logging
import enum
import numpy

from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.numpy import NumpyUtil
from astrophysix.utils.strings import Stringifiable
from astrophysix import units as U
from ..utils import GalacticaValidityCheckMixin
from ...utils import FileType, FileUtil


log = logging.getLogger("astrophysix.simdm")


@enum.unique
class PlotType(enum.Enum):
    """
    Plot type enum

    Example
    -------
    >>> pt = PlotType.HISTOGRAM_2D
    >>> pt.alias
    "2d_hist"
    >>> pt.display_name
    "2D histogram"
    >>> pt.ndimensions
    2
    """
    LINE_PLOT = ("line", "Line plot", 1, 0)
    SCATTER_PLOT = ("scatter", "Scatter plot", 1, 0)
    HISTOGRAM = ("hist", "Histogram", 1, 1)
    HISTOGRAM_2D = ("2d_hist", "2D histogram", 2, 1)
    IMAGE = ("img", "Image", 2, 1)
    MAP_2D = ("2d_map", "2D map", 2, 1)

    def __init__(self, alias, display_name, ndim, aoffset):
        self._alias = alias
        self._display_name = display_name
        self._ndims = ndim
        self._axis_size_offset = aoffset

    @property
    def display_name(self):
        """Plot type verbose name"""
        return self._display_name

    @property
    def ndimensions(self):
        """Plot type number of dimensions"""
        return self._ndims

    @property
    def axis_size_offset(self):
        return self._axis_size_offset

    @property
    def alias(self):
        """Plot type alias"""
        return self._alias

    @classmethod
    def from_alias(cls, alias):
        """
        Find a PlotType according to its alias

        Parameters
        ----------
        alias: :obj:`string`
            required plot type alias

        Returns
        -------
        ft: :class:`~astrophysix.simdm.datafiles.plot.PlotType`
            Plot type matching the requested alias.

        Raises
        ------
        ValueError
            if requested alias does not match any plot type.

        Example
        -------
            >>> pt = PlotType.from_alias("hist")
            >>> pt.display_name
            "Histogram"
            >>> pt2 = PlotType.from_alias("MY_UNKNOWN_PLOT_YPE")
            ValuerError: No PlotType defined with the alias 'MY_UNKNOWN_PLOT_YPE'.
        """
        for t in cls:
            if t.alias == alias:
                return t
        raise ValueError("No PlotType defined with the alias '{a:s}'.".format(a=alias))

    def __str__(self):
        return self._display_name

    def __repr__(self):
        return self.__str__()


class PlotInfo(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Datafile class (Simulation data model)

    Parameters
    ----------
    plot_type: :class:`~astrophysix.simdm.datafiles.plot.PlotType` or :obj:`string`
        Plot type or plot type alias (mandatory)
    xaxis_values: :class:`numpy.ndarray`
        x-axis coordinate values numpy  1D array (mandatory).
    yaxis_values: :class:`numpy.ndarray`
        y-axis coordinate numpy 1D array (mandatory).
    values: :class:`numpy.ndarray`
        plot data values numpy array (mandatory for 2D plots).
    xlabel: :obj:`string`
        x-axis label
    ylabel: :obj:`string`
        y-axis label
    values_label: :obj:`string`
        plot values label
    xaxis_unit:
        TODO
    yaxis_unit:
        TODO
    values_unit:
        TODO
    xaxis_log_scale: :obj:`bool`
        TODO
    yaxis_log_scale: :obj:`bool`
        TODO
    values_log_scale: :obj:`bool`
        TODO
    plot_title: :obj:`string`
        Plot title.
    """

    VALUES_ATTR_NAME = "values"
    VALUES_MD5SUM_ATTR_NAME = "values_md5sum"
    XAXIS_ATTR_NAME = "xaxis_coords"
    XAXIS_MD5SUM_ATTR_NAME = "xaxis_md5sum"
    YAXIS_ATTR_NAME = "yaxis_coords"
    YAXIS_MD5SUM_ATTR_NAME = "yaxis_md5sum"
    PLOT_DATA_JSON_FPATH = "gald3_plot_data.json"

    def __init__(self, **kwargs):
        """

        :param kwargs:
        """
        super(PlotInfo, self).__init__(**kwargs)

        self._type = PlotType.LINE_PLOT
        self._values = None
        self._xaxis = None
        self._yaxis = None
        self._values_md5sum = None
        self._xaxis_md5sum = None
        self._yaxis_md5sum = None

        self._xlabel = ""
        self._ylabel = ""
        self._vlabel = ""
        self._title = ""

        self._xlog = False
        self._ylog = False
        self._vlog = False

        self._xunit = U.none
        self._yunit = U.none
        self._vunit = U.none

        # Set plot type (mandatory)
        if "plot_type" not in kwargs:
            err_msg = "{cname:s} 'plot_type' attribute is not defined (mandatory).".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        ptype = kwargs["plot_type"]
        try:
            # Cast from string value (PlotType alias)
            sptype = Stringifiable.cast_string(ptype)
            self._type = PlotType.from_alias(sptype)
        except ValueError as ve:  # unknown PlotType alias
            log.error(str(ve))
            raise AttributeError(str(ve))
        except TypeError:  # Not a valid string
            if not isinstance(ptype, PlotType):
                err_msg = "{cname:s} 'plot_type' attribute is not a valid PlotType enum " \
                          "value.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._type = ptype

        # Plot title
        if "title" in kwargs:
            self.title = kwargs["title"]

        if not kwargs.get("hdf5_init", False):
            # X-axis coordinates
            if "xaxis_values" not in kwargs:
                err_msg = "{cname:s} 'xaxis_values' attribute is not defined " \
                          "(mandatory).".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

            # y-axis coordinates
            if "yaxis_values" not in kwargs:
                err_msg = "{cname:s} 'yaxis_values' attribute is not defined " \
                          "(mandatory).".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

            # Set data
            self.set_data(kwargs["xaxis_values"], kwargs["yaxis_values"], values=kwargs.get("values", None))

        # X-axis label
        if "xlabel" in kwargs:
            self.xlabel = kwargs["xlabel"]

        # X-axis log-scale flag
        if "xaxis_log_scale" in kwargs:
            self.xaxis_log_scale = kwargs["xaxis_log_scale"]

        # Y-axis log-scale flag + label
        if "yaxis_log_scale" in kwargs:
            self.yaxis_log_scale = kwargs["yaxis_log_scale"]
        if "ylabel" in kwargs:
            self.ylabel = kwargs["ylabel"]

        # 2D plot value log-scale flag + label
        if self._type.ndimensions > 1:
            if "values_log_scale" in kwargs:
                self.values_log_scale = kwargs["values_log_scale"]
            if "values_label" in kwargs:
                self.values_label = kwargs["values_label"]

        # Units
        if "xaxis_unit" in kwargs:
            self.xaxis_unit = kwargs["xaxis_unit"]
        if "yaxis_unit" in kwargs:
            self.yaxis_unit = kwargs["yaxis_unit"]
        if self._type.ndimensions > 1 and "values_unit" in kwargs:
            self.values_unit = kwargs["values_unit"]

    def __eq__(self, other_plot_info):
        """
        PlotInfo comparison method

        Parameters
        ----------
        other_plot_info: :class:`~astrophysix.simdm.datafiles.plot.PlotInfo`
            plot info object to compare to:
        """
        if not super(PlotInfo, self).__eq__(other_plot_info):
            return False

        # Plot type comparison
        if self._type != other_plot_info.plot_type:
            return False

        # Label and title comparison
        if self._title != other_plot_info.title:
            return False

        if self._xlabel != other_plot_info.xlabel:
            return False
        if self._ylabel != other_plot_info.ylabel:
            return False
        if self._type.ndimensions > 1 and self._vlabel != other_plot_info.values_label:
            return False

        # Log-scale flag comparison
        if self._xlog != other_plot_info.xaxis_log_scale:
            return False
        if self._ylog != other_plot_info.yaxis_log_scale:
            return False
        if self._type.ndimensions > 1 and self._vlog != other_plot_info.values_log_scale:
            return False

        # Unit comparison
        if not self._xunit.identical(other_plot_info.xaxis_unit):
            return False
        if not self._yunit.identical(other_plot_info.yaxis_unit):
            return False
        if self._type.ndimensions > 1 and not self._vunit.identical(other_plot_info.values_unit):
            return False

        # MD5 sum comparison
        if self._xaxis_md5sum != other_plot_info._xaxis_md5sum or self._yaxis_md5sum != other_plot_info._yaxis_md5sum:
            return False

        if self._type.ndimensions > 1 and self._values_md5sum != other_plot_info._values_md5sum:
            return False

        return True

    @property
    def plot_type(self):
        """Returns the plot type (:class:`~astrophysix.simdm.datafiles.plot.PlotType`). Cannot be edited."""
        return self._type

    @property
    def title(self):
        """Plot title. Can be set to any :obj:`string` value."""
        return self._title

    @title.setter
    def title(self, new_title):
        try:
            self._title = Stringifiable.cast_string(new_title, valid_empty=True)
        except TypeError:
            err_msg = "{cname:s} 'title' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def set_data(self, xaxis_values, yaxis_values, values=None):
        """
        Set plot data arrays.

        Parameters
        ----------
        xaxis_values: :class:`numpy.ndarray`
            x-axis coordinate array
        yaxis_values: :class:`numpy.ndarray`
            TODO
        values: :class:`numpy.ndarray`
            TODO
        """
        # X-axis plot coordinates
        try:
            NumpyUtil.check_is_array(xaxis_values, ndim=1)
        except AttributeError:
            err_msg = "'xaxis_values' {cname:s} attribute must be a 1-dimensional " \
                      "array.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # x-axis coordinate array size
        nx = xaxis_values.size

        # Y-axis plot coordinates
        try:
            NumpyUtil.check_is_array(yaxis_values, ndim=1)
        except AttributeError:
            err_msg = "'yaxis_values' {cname:s} attribute must be a 1-dimensional " \
                      "array.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # Check y-axis coordinate array size
        ny = yaxis_values.size
        if self._type.ndimensions == 1:  # 1D plots
            ny_target = nx - self._type.axis_size_offset
            if ny != ny_target:
                raise AttributeError("Array size mismatch : 'yaxis_values' coordinate array size should be {siz:d} "
                                     "(x-axis coordinate array size={nx:d}) for '{pt!s}'.".format(siz=ny_target, nx=nx,
                                                                                                  pt=self._type))

        # Plot values
        if self._type.ndimensions > 1:
            if values is None:
                err_msg = "{cname:s} 'values' attribute is not defined (mandatory for 2D " \
                          "plots).".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

            try:
                NumpyUtil.check_is_array(values, self._type.ndimensions)
            except AttributeError:
                raise AttributeError("{cname:s} 'values' attribute must be a {nd:d}-dimensional "
                                     "array.".format(cname=self.__class__.__name__, nd=self._type.ndimensions))

            nvx = nx - self._type.axis_size_offset
            nvy = ny - self._type.axis_size_offset
            if values.shape[0] != nvx or values.shape[1] != nvy:
                raise AttributeError("Array size mismatch : 'values' array (shape={sht!s}) should have a shape "
                                     "({nvx:d}, {nvy:d}) (x-axis coordinate array size={nx:d} ; y-axis coordinate "
                                     "array size={ny:d}) for '{pt!s}'.".format(sht=values.shape, nvx=nvx, nvy=nvy,
                                                                               nx=nx, ny=ny, pt=self._type))

        # ------------------------------------------ Set new data ---------------------------------------------------- #
        # Compute the md5 checksum of a the new value array
        self._xaxis = xaxis_values
        self._xaxis_md5sum = NumpyUtil.md5sum(xaxis_values)
        self._yaxis_md5sum = NumpyUtil.md5sum(yaxis_values)
        self._yaxis = yaxis_values
        self._values = values if self._type.ndimensions > 1 else None
        self._values_md5sum = NumpyUtil.md5sum(values) if self._type.ndimensions > 1 else None

        # Flag the PlotInfo object as 'loaded in memory' to avoid reading data from HDF5 study file in the future
        self._hsp_set_lazy_read()
        # ------------------------------------------------------------------------------------------------------------ #

    @property
    def xaxis_values(self):
        """
        Plot x-axis coordinate array (:class:`numpy.ndarray`). Cannot be edited. Implements lazy I/O.

        Note
        ----
            To edit plot values, see :func:`PlotInfo.set_data` method.
        """
        self._hsp_lazy_read()
        return self._xaxis

    @property
    def yaxis_values(self):
        """
        Plot y-axis coordinate array (:class:`numpy.ndarray`). Cannot be edited. Implements lazy I/O.

        Note
        ----
            To edit plot values, see :func:`PlotInfo.set_data` method.
        """
        self._hsp_lazy_read()
        return self._yaxis

    @property
    def values(self):
        """
        Plot values array. Cannot be edited. Implements lazy I/O.

        Note
        ----
            To edit plot values, see :func:`PlotInfo.set_data` method.
        """
        if self._type.ndimensions > 1:
            self._hsp_lazy_read()
            return self._values
        else:
            raise AttributeError("{cname:s} object does not have a 'values' "
                                 "property.".format(cname=self.__class__.__name__))

    @property
    def xlabel(self):
        """x-axis label. Can be set to any :obj:`string` value."""
        return self._xlabel

    @xlabel.setter
    def xlabel(self, new_xaxis_label):
        try:
            self._xlabel = Stringifiable.cast_string(new_xaxis_label, valid_empty=True)
        except TypeError:
            err_msg = "{cname:s} 'xlabel' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def ylabel(self):
        """y-axis label. Can be set to any :obj:`string` value."""
        return self._ylabel

    @ylabel.setter
    def ylabel(self, new_yaxis_label):
        try:
            self._ylabel = Stringifiable.cast_string(new_yaxis_label, valid_empty=True)
        except TypeError:
            err_msg = "{cname:s} 'ylabel' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def values_label(self):
        """plot values label. Can be set to any :obj:`string` value."""
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_label' "
                                 "property.".format(cname=self.__class__.__name__))
        return self._vlabel

    @values_label.setter
    def values_label(self, new_vlabel):
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_label' "
                                 "property.".format(cname=self.__class__.__name__))

        try:
            self._vlabel = Stringifiable.cast_string(new_vlabel, valid_empty=True)
        except TypeError:
            err_msg = "{cname:s} 'values_label' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def xaxis_unit(self):
        """TODO"""
        return self._xunit

    @xaxis_unit.setter
    def xaxis_unit(self, new_xaxis_unit):
        if isinstance(new_xaxis_unit, U.Unit):
            self._xunit = new_xaxis_unit
        else:
            try:
                s = Stringifiable.cast_string(new_xaxis_unit, valid_empty=False)
                self._xunit = U.Unit.from_name(s)
            except TypeError:  # Not a valid string
                err_msg = "{cname:s} 'xaxis_unit' property is not a valid (non-empty) " \
                          "string.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            except AttributeError as aerr:
                err_msg = "{cname:s} 'xaxis_unit' property error : {uerr:s}.".format(uerr=str(aerr),
                                                                                     cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

    @property
    def yaxis_unit(self):
        """TODO"""
        return self._yunit

    @yaxis_unit.setter
    def yaxis_unit(self, new_yaxis_unit):
        if isinstance(new_yaxis_unit, U.Unit):
            self._yunit = new_yaxis_unit
        else:
            try:
                s = Stringifiable.cast_string(new_yaxis_unit, valid_empty=False)
                self._yunit = U.Unit.from_name(s)
            except TypeError:  # Not a valid string
                err_msg = "{cname:s} 'yaxis_unit' property is not a valid (non-empty) " \
                          "string.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            except AttributeError as aerr:
                err_msg = "{cname:s} 'yaxis_unit' property error : {uerr:s}.".format(uerr=str(aerr),
                                                                                     cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

    @property
    def values_unit(self):
        """TODO"""
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_unit' "
                                 "property.".format(cname=self.__class__.__name__))

        return self._vunit

    @values_unit.setter
    def values_unit(self, new_vunit):
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_unit' "
                                 "property.".format(cname=self.__class__.__name__))

        if isinstance(new_vunit, U.Unit):
            self._vunit = new_vunit
        else:
            try:
                s = Stringifiable.cast_string(new_vunit, valid_empty=False)
                self._vunit = U.Unit.from_name(s)
            except TypeError:  # Not a valid string
                err_msg = "{cname:s} 'values_unit' property is not a valid (non-empty) " \
                          "string.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            except AttributeError as aerr:
                err_msg = "{cname:s} 'values_unit' property error : {uerr:s}.".format(uerr=str(aerr),
                                                                                      cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

    @property
    def xaxis_log_scale(self):
        """x-axis log scale boolean flag. Can be edited to any :obj:`bool` value."""
        return self._xlog

    @xaxis_log_scale.setter
    def xaxis_log_scale(self, new_log):
        if not isinstance(new_log, bool):
            err_msg = "'xaxis_log_scale' {cname:s} property must be a boolean " \
                      "value.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._xlog = new_log

    @property
    def yaxis_log_scale(self):
        """y-axis log scale boolean flag. Can be edited to any :obj:`bool` value."""
        return self._ylog

    @yaxis_log_scale.setter
    def yaxis_log_scale(self, new_log):
        if not isinstance(new_log, bool):
            err_msg = "'yaxis_log_scale' {cname:s} property must be a boolean " \
                      "value.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._ylog = new_log

    @property
    def values_log_scale(self):
        """value log scale boolean flag. Can be edited to any :obj:`bool` value."""
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_log_scale' "
                                 "property.".format(cname=self.__class__.__name__))
        return self._vlog

    @values_log_scale.setter
    def values_log_scale(self, new_log):
        if self._type.ndimensions < 2:
            raise AttributeError("{cname:s} object does not have a 'values_log_scale' "
                                 "property.".format(cname=self.__class__.__name__))

        if not isinstance(new_log, bool):
            err_msg = "'values_log_scale' {cname:s} property must be a boolean " \
                      "value.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._vlog = new_log

    def _export_to_json(self, fpath):
        """
        Export plot information to a dict and dump it into a JSON file
        :param fpath: JSON file path
        """
        json_data = {"type": self._type.alias, "title": self._title, "xaxis_log": self._xlog, "yaxis_log": self._ylog,
                     "xaxis_label": self._xlabel, "yaxis_label": self._ylabel,
                     "xaxis_unit": self._xunit.latex, "yaxis_unit": self._yunit.latex}
        if self._type == PlotType.HISTOGRAM_2D:
            json_data.update({"xaxis_coordinates": list(self._xaxis), "yaxis_coordinates": list(self._yaxis)})
            vals = []
            mnan = numpy.isnan(self._values)
            for i in range(self._xaxis.size - self._type.axis_size_offset):
                for j in range(self._yaxis.size - self._type.axis_size_offset):
                    if not mnan[i, j]:
                        vals.append([i, j, self._values[i, j]])
            json_data["values"] = vals
            json_data["value_log"] = self._vlog
            json_data["value_label"] = self._vlabel
            json_data["value_unit"] = self._vunit.latex
        elif self._type == PlotType.HISTOGRAM:
            json_data.update({"xaxis_coordinates": list(self._xaxis), "values": list(self._yaxis)})
        elif self._type == PlotType.LINE_PLOT:
            json_data.update({"xaxis_coordinates": list(self._xaxis), "values": list(self._yaxis)})
        else:
            raise NotImplementedError()

        with open(fpath, "w") as f:
            json.dump(json_data, f)
        log.info("File '{fp:s}' saved.".format(fp=fpath))

    def save_json_plot_data(self, json_filepath):
        """
        Save plot information into an external JSON file on the local filesystem

        Parameters
        ----------
        json_filepath: :obj:`string`
            valid external JSON file path
        """
        file_path = Stringifiable.cast_string(json_filepath, valid_empty=False)
        fpath = FileUtil.new_filepath(file_path, FileType.JSON_FILE, create_dir=True)
        ftype = FileUtil.get_file_type(json_filepath)
        if ftype != FileType.JSON_FILE:
            err_msg = "Invalid filename for a {ft:s} file ({fp!s}).".format(ft=FileType.JSON_FILE.alias,
                                                                            fp=json_filepath)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # Force file data loading if not already read from HDF5 file
        self._hsp_lazy_read()

        # Do the actual file saving operation...
        self._export_to_json(fpath)

        return fpath

    def gald3_render(self, open_browser=True, local_http_port=8076):
        """
        Exports the PlotInfo object into a temporary JSON file and opens a web browser window to render it using the
        Galactica GalD3.js javascript library.

        Parameters
        ----------
        open_browser: : obj:`bool`
            open a web browser window. Default True. Set to False if already opened
        local_http_port: :obj:`int`
            localhost http port to use in the URL
        """
        f, tfpath = tempfile.mkstemp(prefix="d3_temp_plot_data_", suffix=".json", text=True)
        self.save_json_plot_data(tfpath)
        tmp_http_fpath = os.path.join(tempfile.gettempdir(), PlotInfo.PLOT_DATA_JSON_FPATH)
        if os.path.islink(tmp_http_fpath):
            old_plot_data_file = os.readlink(tmp_http_fpath)
            log.info("Deleting link to temporary JSON data file '{fp:s}'.".format(fp=old_plot_data_file))
            os.remove(tmp_http_fpath)
            os.remove(old_plot_data_file)

        log.info("Creating new symlink to temporary JSON data file '{fp:s}'.".format(fp=tfpath))
        os.symlink(tfpath, tmp_http_fpath)
        log.info("Opening browser to url : http://127.0.0.1:{p:d}/chart".format(p=local_http_port))
        if open_browser:
            webbrowser.open("http://127.0.0.1:{p:d}/chart".format(p=local_http_port))

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["plot_type", "title", "hdf5_init", "xaxis_values", "yaxis_values", "values", "xlabel", "ylabel",
                "xaxis_log_scale", "yaxis_log_scale", "values_log_scale", "values_label",
                "xaxis_unit", "yaxis_unit", "values_unit"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a PlotInfo object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the PlotInfo into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(PlotInfo, self)._hsp_write(h5group, **kwargs)

        # -------------------------------- Write datafile plot info info to HDF5 ------------------------------------- #
        # Write plot type
        self._hsp_write_attribute(h5group, ("plot_type", self._type.alias), **kwargs)

        # Write labels (x-axis/y-axis/values/title)
        if self._type.ndimensions > 1:
            self._hsp_write_attribute(h5group, ("vlabel", self._vlabel), **kwargs)
        self._hsp_write_attribute(h5group, ("xlabel", self._xlabel), **kwargs)
        self._hsp_write_attribute(h5group, ("ylabel", self._ylabel), **kwargs)
        self._hsp_write_attribute(h5group, ("title", self._title), **kwargs)

        # Write log-scale flags (x-axis/y-axis/value)
        self._hsp_write_attribute(h5group, ("xlog", self._xlog), **kwargs)
        self._hsp_write_attribute(h5group, ("ylog", self._ylog), **kwargs)
        if self._type.ndimensions > 1:
            self._hsp_write_attribute(h5group, ("vlog", self._vlog), **kwargs)

        # Write units
        self._hsp_write_attribute(h5group, ("xunit", self._xunit), **kwargs)
        self._hsp_write_attribute(h5group, ("yunit", self._yunit), **kwargs)
        if self._type.ndimensions > 1:
            self._hsp_write_attribute(h5group, ("vunit", self._vunit), **kwargs)

        # Write array MD5 sums
        write_values = False
        if self._type.ndimensions > 1:
            values_attribute_name = "_values"  # => self._values
            write_values = self._hsp_write_dataset(h5group, (self.VALUES_ATTR_NAME, values_attribute_name),
                                                   md5sum_params=(self.VALUES_MD5SUM_ATTR_NAME, self._values_md5sum),
                                                   **kwargs)
        xaxis_attribute_name = "_xaxis"  # => self._xaxis
        write_xaxis = self._hsp_write_dataset(h5group, (self.XAXIS_ATTR_NAME, xaxis_attribute_name),
                                              md5sum_params=(self.XAXIS_MD5SUM_ATTR_NAME, self._xaxis_md5sum),
                                              **kwargs)
        yaxis_attribute_name = "_yaxis"  # => self._yaxis
        write_yaxis = self._hsp_write_dataset(h5group, (self.YAXIS_ATTR_NAME, yaxis_attribute_name),
                                              md5sum_params=(self.YAXIS_MD5SUM_ATTR_NAME, self._yaxis_md5sum),
                                              **kwargs)

        if write_values or write_xaxis or write_yaxis:
            self._hsp_write_callback(str(self), **kwargs)
        # ------------------------------------------------------------------------------------------------------------ #

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a PlotInfo object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        pi: ``PlotInfo``
            Read PlotInfo instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(PlotInfo, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read plot type and title
        title = cls._hsp_read_attribute(h5group, "title", "plot title")
        ptype = PlotType.from_alias(cls._hsp_read_attribute(h5group, "plot_type", "plot type"))

        # Read labels (x-axis/y-axis/values/title)
        xlabel = cls._hsp_read_attribute(h5group, "xlabel", "x-axis label")
        ylabel = cls._hsp_read_attribute(h5group, "ylabel", "y-axis label")
        kw = {"uid": uid, "plot_type": ptype, "title": title, "xlabel": xlabel, "ylabel": ylabel, "hdf5_init": True}
        if ptype.ndimensions > 1:
            vlabel = cls._hsp_read_attribute(h5group, "vlabel", "plot values label")
            if vlabel is not None:
                kw["values_label"] = vlabel

        # Read log-scale flags
        kw["xaxis_log_scale"] = bool(cls._hsp_read_attribute(h5group, "xlog", "x-axis log-scale flag"))
        kw["yaxis_log_scale"] = bool(cls._hsp_read_attribute(h5group, "ylog", "y-axis log-scale flag"))
        if ptype.ndimensions > 1:
            vlog = cls._hsp_read_attribute(h5group, "vlog", "values log-scale flag")
            if vlog is not None:
                kw["values_log_scale"] = bool(vlog)

        # Read units
        kw["xaxis_unit"] = cls._hsp_read_unit(h5group, "xunit")
        kw["yaxis_unit"] = cls._hsp_read_unit(h5group, "yunit")
        if ptype.ndimensions > 1:
            kw["values_unit"] = cls._hsp_read_unit(h5group, "vunit")

        pinfo = cls(**kw)

        # Read x-axis/y-axis/value array md5sum
        if ptype.ndimensions > 1:
            pinfo._values_md5sum = pinfo._hsp_read_attribute(h5group, cls.VALUES_MD5SUM_ATTR_NAME,
                                                             "plot info value array md5sum",
                                                             raise_error_if_not_found=True)
        pinfo._xaxis_md5sum = pinfo._hsp_read_attribute(h5group, cls.XAXIS_MD5SUM_ATTR_NAME,
                                                        "plot info x-axis coordinate array md5sum",
                                                        raise_error_if_not_found=True,)
        pinfo._yaxis_md5sum = pinfo._hsp_read_attribute(h5group, cls.YAXIS_MD5SUM_ATTR_NAME,
                                                        "plot info y-axis coordinate/value array md5sum",
                                                        raise_error_if_not_found=True)

        # Set HDF5 group/file info for lazy I/O
        pinfo._hsp_set_lazy_source(h5group)

        return pinfo

    def _hsp_lazy_read_data(self, h5group):
        """
        Lazy read method to load plot values + PIL image from HDF5 file (*.h5)

        Parameters
        ----------
        h5group: `h5py.Group`
        """
        # Read coord/value arrays
        if self._type.ndimensions > 1:
            self._values = self._hsp_read_dataset(h5group, self.VALUES_ATTR_NAME, "plot info value array",
                                                  raise_error_if_not_found=True)
        self._xaxis = self._hsp_read_dataset(h5group, self.XAXIS_ATTR_NAME, "plot info x-axis coordinate array",
                                             raise_error_if_not_found=True)
        self._yaxis = self._hsp_read_dataset(h5group, self.YAXIS_ATTR_NAME, "plot info y-axis coordinate/value array",
                                             raise_error_if_not_found=True)

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        # TODO Galactica validity check for PlotInfo
        pass

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = "[{tname:s}] plot information".format(tname=self._type.name)
        return s


__all__ = ["PlotType", "PlotInfo"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
Datatype enum
-------------

.. autoclass:: astrophysix.simdm.utils.DataType
    :members:
    :undoc-members:


Object lists
------------

.. autoclass:: astrophysix.simdm.utils.ObjectList
   :members:
   :no-special-members: __call__
   :exclude-members: add_deletion_handler

"""
from __future__ import unicode_literals  # Python 2 and 3 compatibility
from future.builtins import str, list
import enum
import re
import logging
import numpy as N

from astrophysix.utils import Stringifiable

log = logging.getLogger("astrophysix.simdm")


class DataType(enum.Enum):
    """
    Value data type enum

    Example
    -------
    >>> dt = DataType.INTEGER
    >>> dt.name
    "Integer number"

    """
    BOOLEAN = ('bool', "Boolean")
    COMPLEX = ('comp', "Complex number")
    DATETIME = ('time', "Datetime")
    REAL = ('real', "Real number")
    INTEGER = ('int', "Integer number")
    RATIONAL = ('rat', "Rational number")
    STRING = ('str', "String")

    def __init__(self, key, name):
        self._key = key
        self._name = name

    @property
    def key(self):
        """Data type index key"""
        return self._key

    @property
    def name(self):
        """Data type verbose name"""
        return self._name

    @classmethod
    def from_key(cls, k):
        """
        Parameters
        ----------
        key: :obj:`string`
            data type key

        Returns
        -------
        t: :class:`~astrophysix.simdm.utils.DataType`
            Physics matching the requested key.

        Raises
        ------
        ValueError
            if requested key does not match any physics.

        Example
        -------
            >>> dt = DataType.from_key("rat")
            >>> dt.name
            "Rational number"
            >>> dt2 = DataType.from_key("MY_UNKNOWN_DTYPE")
            ValuerError: No DataType defined with the key 'MY_UNKNOWN_DTYPE'.
        """
        for dt in cls:
            if dt.key == k:
                return dt
        raise ValueError("No DataType defined with the key '{key:s}'.".format(key=k))


class GalacticaValidityCheckMixin(object):
    INVALID_ALIAS_ERROR_MESSAGE = "The alias can contain capital letters, digits and \'_\' only. It must start with " \
                                  "a capital letter and cannot end with a \'_\'."
    VALID_ALIAS_REGEX = re.compile("""^[A-Z]        # First character must be a capital letter
                                      ([A-Z0-9_]*   # Then any capital letter, digit or '_'
                                      [A-Z0-9])?$   # Must ends with a capital letter or a digit""", re.VERBOSE)

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        pass

    def galactica_valid_alias(self, alias_value):
        if self.VALID_ALIAS_REGEX.match(alias_value) is None:
            return self.INVALID_ALIAS_ERROR_MESSAGE
        return None


class ObjectList(GalacticaValidityCheckMixin, Stringifiable):
    """
    Generic object list container class

    Parameters
    ----------
    obj_class: :obj:`type`
        base class of the objects that can be added to the list
    index_prop_name: :obj:`string`
        object property name used as a list index
    object_addition_vcheck: :obj:`tuple` (:obj:`callable` validity check method, :obj:`string` added object list property name)
        validity check method to install into an object's object list upon object addition into the list. Default None.
    object_addition_delhandler: :obj:`tuple` (:obj:`callable` deletion handler, :obj:`string` added object list property name)
        deletion handler to install into an object's object list upon object addition into the list. Default None.

    Examples
    --------
        >>> run1 = Simulation(simu_code=arepo, name="Pure-hydro run (isolated galaxy)")
        >>> run2 = Simulation(simu_code=arepo, name="MHD run")
        >>> run3 = project.simulation.add(Simulation(simu_code=arepo, name="Hydro run with BH feedback")
        >>> run4 = Simulation(simu_code=arepo, name="MHD run with BH feedback")
        >>> project.simulation.add(run1)
        >>> project.simulation.add(run2)
        >>> project.simulation.add(run3)
        >>> project.simulation.add(run4, insert_pos=2)  # Insert at position 2, not appendend at the end of the list
        >>> len(project.simulations)
        4
        >>> print(str(project.simulations))
        Simulation list :
        +---+-----------------------------------+-----------------------------------------------+
        | # |              Index                |                          Item                 |
        +---+-----------------------------------+-----------------------------------------------+
        | 0 | Pure-hydro run (isolated galaxy)  | 'Pure-hydro run (isolated galaxy)' simulation |
        +---+-----------------------------------+-----------------------------------------------+
        | 1 | MHD run                           | 'MHD run' simulation                          |
        +---+-----------------------------------+-----------------------------------------------+
        | 2 | MHD run with BH feedback          | 'MHD run with BH feedback' simulation         |
        +---+-----------------------------------+-----------------------------------------------+
        | 3 | Hydro run with BH feedback        | 'Hydro run with BH feedback' simulation       |
        +---+-----------------------------------+-----------------------------------------------+
        >>> run3 is project.simulations[3]  # Search by item position
        True
        >>> project.simulations["MHD run"]  # Search by item index value
        'MHD run' simulation
        >>> del project.simulations[0]
        >>> del project.simulations["MHD run"]
        >>> del project.simulations[run4]
        >>> print(str(project.simulations))
        Simulation list :
        +---+-----------------------------------+-----------------------------------------------+
        | # |              Index                |                          Item                 |
        +---+-----------------------------------+-----------------------------------------------+
        | 0 | Hydro run with BH feedback        | 'Hydro run with BH feedback' simulation       |
        +---+-----------------------------------+-----------------------------------------------+
    """
    def __init__(self, obj_class, index_prop_name, object_addition_vcheck=None, object_addition_delhandler=None):
        super(ObjectList, self).__init__()
        self._list = list()
        self._obj_class = obj_class
        self._index_prop_name = index_prop_name
        self._validity_check_methods = [self._index_unicity_validity_check]
        self._deletion_handlers = []
        if object_addition_vcheck is not None:
            self._propag_obj_add_vc_method, self._added_obj_list_prop_name = object_addition_vcheck
        else:
            self._propag_obj_add_vc_method = None
            self._added_obj_list_prop_name = ""
        if object_addition_delhandler is not None:
            self._propag_obj_add_delhandler, self._added_obj_list_prop_name_dh = object_addition_delhandler
        else:
            self._propag_obj_add_delhandler = None
            self._added_obj_list_prop_name_dh = ""

    def __eq__(self, other):
        """
        Object list comparison method

        Parameters
        ----------
        other: :class:`~astrophysix.simdm.utils.ObjectList`
            other object list to compare to
        """
        # Object classes differ => not equal
        if self._obj_class != other.object_class:
            return False

        # Indexing property differ => not equal
        if self._index_prop_name != other.index_attribute_name:
            return False

        # List lengths differ => not equal
        if len(other) != len(self._list):
            return False

        # Check each object equality (both lists DO have the same length here)
        for iobj, o in enumerate(other):
            if o != self._list[iobj]:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        """Size of the object list"""
        return len(self._list)

    @property
    def index_attribute_name(self):
        """
        Name of the object property used as an index in this object list
        """
        return self._index_prop_name

    @property
    def object_class(self):
        """
        Type of object that can be added into the list
        """
        return self._obj_class

    def _index_unicity_validity_check(self, obj):
        # Check unicity of object with this given index attribute value within the list
        index_val = getattr(obj, self._index_prop_name, None)
        if index_val is not None and self.__contains__(index_val):
            err_msg = "Cannot add {cname:s} object with index '{iv!s}' in this list, another item with that index " \
                      "value already exists.".format(cname=self._obj_class.__name__, iv=index_val)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def add_validity_check_method(self, can_add_meth):
        """
        Add an object addition validity check method to the list of addition validity check methods

        Parameters
        ----------
        can_add_meth: ``Callable``
            object addition validity check method
        """
        # Add addition validity check method to the list of object list addition validity check methods
        if can_add_meth not in self._validity_check_methods:  # Warning here method __eq__() comparison method will be called
            self._validity_check_methods.append(can_add_meth)

    def _can_add_object(self, obj):
        if not isinstance(obj, self._obj_class):
            err_msg = "Added object is not a valid '{cname:s}' instance.".format(cname=self._obj_class.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # ---------------------------------------- Validity check methods ------------------------------------------- #
        # Call validity check methods, if any is defined
        for validity_check_meth in self._validity_check_methods:
            validity_check_meth(obj)

    def add_deletion_handler(self, can_delete_meth):
        """
        Add an object deletion handling method to the list of deletion handlers

        Parameters
        ----------
        can_delete_meth: :obj:`callable`
            object deletion handling method
        """
        # Add deletion handling method to the list of object list deletion handlers
        if can_delete_meth not in self._deletion_handlers:  # Warning here hendling instance __eq__() comparison method will be called
            self._deletion_handlers.append(can_delete_meth)

    def __getitem__(self, index):
        """
        Get an object from the list.

        Parameters
        ----------
        item: :obj:`int` or :obj:`string`
            object position in the list (:obj:`int`) or index property value (:obj:`string`) of the object
            to fetch from the list.

        Returns
        -------
        o: :obj:`object` of type self.object_class
            Found object in the list. None if none were found.

        Raises
        ------
        AttributeError
            if the search index type is neither an :obj:`int` nor a :obj:`string`.
        IndexError
            if the :obj:`int` search index value is lower than 0 or larger than the length of the list - 1.
        """
        if Stringifiable.is_type_string(index):
            sindex = Stringifiable.cast_string(index)
            for item in self._list:
                index_val = getattr(item, self._index_prop_name, None)
                if index_val is not None and index_val == sindex:
                    return item

            log.warning("Cannot find '{idx!s}' {cln:s} instance in list !".format(idx=sindex,
                                                                                  cln=self._obj_class.__name__))
            return None
        elif type(index) == int:
            if index >= 0 and index < len(self._list):
                return self._list[index]
            err_msg = "Object list index out of range (len={l:d}).".format(l=len(self._list))
            log.error(err_msg)
            raise IndexError(err_msg)
        else:
            err_msg = "'{it:s}' is not a valid search index. Valid types are 'str' and " \
                      "'int'.".format(it=str(index), cln=self._obj_class.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def __contains__(self, item):
        if Stringifiable.is_type_string(item):  # Check instance index is in list
            sitem = Stringifiable.cast_string(item)
            for obj in self._list:
                index_val = getattr(obj, self._index_prop_name, None)
                if index_val is not None and index_val == sitem:
                    return True
            return False
        elif isinstance(item, self._obj_class):  # Check instance is in list
            if item in self._list:
                return True
            return False
        else:
            err_msg = "'{it!s}' is not a valid search index. Valid types are 'str' and '{cln:s}' " \
                      "objects.".format(it=item, cln=self._obj_class.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def add(self, obj, insert_pos=-1):
        """
        Adds a instance to the list at a given position

        Parameters
        ----------
        obj: :obj:`object`
            instance to insert in the list
        insert_pos: :obj:`int`
            insertion position in the simulation list. Default -1 (last).
        """
        self._can_add_object(obj)

        if insert_pos == -1:
            self._list.append(obj)
        else:
            self._list.insert(insert_pos, obj)

        # Propagate validity check method into the added object's internal object list
        if self._propag_obj_add_vc_method is not None and hasattr(obj, self._added_obj_list_prop_name):
            obj_list = getattr(obj, self._added_obj_list_prop_name)
            obj_list.add_validity_check_method(self._propag_obj_add_vc_method)

        # Propagate deletion handler into the added object's internal object list
        if self._propag_obj_add_delhandler is not None:
            if isinstance(self._added_obj_list_prop_name_dh, list):
                target = None
                for prop in self._added_obj_list_prop_name_dh:
                    if target is None and hasattr(obj, prop):
                        target = getattr(obj, prop)
                    elif target is not None and hasattr(target, prop):
                        target = getattr(target, prop)
            elif hasattr(obj, self._added_obj_list_prop_name_dh):
                target = getattr(obj, self._added_obj_list_prop_name_dh)

            obj_list = target
            obj_list.add_deletion_handler(self._propag_obj_add_delhandler)

        return obj

    def find_by_uid(self, uid):
        """
        Find an object in the list with a matching UUID

        Parameters
        ----------
        uid: :obj:`UUID` or :obj:`string`
            UUID or UUID string representation of the object to search for.

        Returns
        -------
        o: Matching object with corresponding UUID,if any. Otherwise returns None
        """
        suid = str(uid)
        if getattr(self._obj_class, "uid", None) is None:
            err_msg = "{cname} objects do not have a 'uid' property.".format(cname=self._obj_class.__name__)
            log.error(err_msg)
            raise TypeError(err_msg)

        for o in self._list:
            if str(o.uid) == suid:
                return o
        return None

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Perform Galactica validity check on list items
        if issubclass(self._obj_class, GalacticaValidityCheckMixin):
            for item in self._list:
                item.galactica_validity_check(**kwargs)

        # Check unicity of object with this given index attribute value within the list
        index_dict = {}
        for item in self._list:
            index_val = getattr(item, self._index_prop_name, None)
            if index_val is not None:
                if index_val in index_dict:
                    log.warning("{o1!s} and {o2!s} share the same '{p:s}' index value in this "
                                "list.".format(o1=item, o2=index_dict[index_val], p=self._index_prop_name))
                else:
                    index_dict[index_val] = item

    def __unicode__(self):
        """
        String representation of the instance
        """
        # Empty object list
        if len(self._list) == 0:
            return "{obj_cname:s} list : empty".format(obj_cname=self._obj_class.__name__)

        # Display object list in a pretty-formatted table
        strrep = "{obj_cname:s} list :\n".format(obj_cname=self._obj_class.__name__)
        obj_slist = list([str(obj) for obj in self])
        index_slist = list([getattr(obj, self._index_prop_name, "") for obj in self])
        npos = int(N.log10(len(obj_slist))) + 1
        nind = N.max([len(ind) for ind in index_slist])
        ind_header = "Index"
        nind = len(ind_header) if nind < len(ind_header) else nind
        nstr = N.max([len(sobj) for sobj in obj_slist])
        item_header = "Item"
        nstr = len(item_header) if nstr < len(item_header) else nstr
        interline = "+-{npos:s}-+-{nind:s}-+-{nstr:s}-+".format(npos="-"*npos, nind=nind*"-", nstr=nstr*"-")
        strrep += interline + "\n| {d:^{npos}s} | {ind:^{nind}s} | " \
                              "{s:^{nstr}s} |\n".format(npos=npos, d="#", nind=nind, ind=ind_header, nstr=nstr,
                                                        s=item_header)
        for i in range(len(obj_slist)):
            strrep += interline + "\n"
            strrep += "| {i:>{npos}d} | {ind:<{nind}s} | {s:<{nstr}s} |\n".format(npos=npos, i=i, nind=nind,
                                                                                  ind=index_slist[i], nstr=nstr, s=obj_slist[i])
        strrep += interline

        return strrep

    def __iter__(self):
        """Basic object list iterator"""
        return iter(self._list)

    def __call__(self, *args, **kwargs):
        return self.__iter__()

    def __delitem__(self, item):
        """
        Delete an object from the list.

        Parameters
        ----------
        item: :obj:`object` or :obj:`int` or :obj:`string`
            instance to delete, object position in the list (:obj:`int`) or index property value (:obj:`string`) of the
            object to remove from the list.
        """
        found_obj = None
        if isinstance(item, self._obj_class):  # item is a corresponding instance of the list object class
            if item in self._list:
                found_obj = item
            else:
                err_msg = "'{o!s}' does not belong to this '{cln:s}' list.".format(o=item, cln=self._obj_class.__name__)
                log.error(err_msg)
                raise KeyError(err_msg)
        elif type(item) == int:
            if item >= 0 and item < len(self._list):
                found_obj = self._list[item]
            else:
                err_msg = "Object list index out of range (len={l:d}).".format(l=len(self._list))
                log.error(err_msg)
                raise IndexError(err_msg)
        elif Stringifiable.is_type_string(item):  # item is a string => search for an object in the list
            sitem = Stringifiable.cast_string(item)
            for obj in self._list:
                index_val = getattr(obj, self._index_prop_name, None)
                if index_val is not None and index_val == sitem:
                    found_obj = obj
                    break

            if found_obj is None:
                # Not found
                err_msg = "Cannot find '{it!s}' {cln:s} instance in list !".format(it=sitem, cln=self._obj_class.__name__)
                log.error(err_msg)
                raise KeyError(err_msg)
        else:
            # Invalid item value
            err_msg = "'{it!s}' is not a valid deletion index. Valid types are 'str' and '{cln:s}' " \
                      "objects.".format(it=item, cln=self._obj_class.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # ------------------------------------------ Checks dependencies --------------------------------------------- #
        # Call object deletion handlers, if any is defined
        depend_list = []
        for del_handler in self._deletion_handlers:
            depend_obj = del_handler(found_obj)
            if depend_obj is not None:
                if isinstance(depend_obj, list):
                    depend_list += depend_obj
                else:
                    depend_list.append(depend_obj)

        # If deleted object has any dependency, prevent its deletion
        if len(depend_list) > 0:
            err_msg = "'{o!s}' cannot be deleted, the following items depend on it (try to delete them first) : " \
                      "[{dl:s}].".format(o=found_obj, dl=", ".join(depend_list))
            log.warning(err_msg)
            raise AttributeError(err_msg)

        self._list.remove(found_obj)

    def clear(self):
        """Clear the object list"""
        for i in range(len(self._list)):
            del self[0]


__all__ = ["ObjectList", "DataType", "GalacticaValidityCheckMixin"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
Parameter visibility flag
^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: astrophysix.simdm.experiment.ParameterVisibility
   :members:
   :undoc-members:


Parameter setting
^^^^^^^^^^^^^^^^^

.. autoclass:: astrophysix.simdm.experiment.ParameterSetting
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX
"""
from __future__ import absolute_import, unicode_literals
from future.builtins import str, list, dict, int
import logging
import uuid

from astrophysix.utils.strings import Stringifiable
from astrophysix.utils.persistency import Hdf5StudyPersistent
from ..utils import DataType, GalacticaValidityCheckMixin
from ..protocol import InputParameter
from astrophysix import units as U
from enum import Enum


log = logging.getLogger("astrophysix.simdm")


class ParameterVisibility(Enum):
    """
    Parameter setting visibility flag (enum)

    Example
    -------
        >>> vis = ParameterVisibility.BASIC_DISPLAY
        >>> vis.display_name
        "Basic display"
    """
    NOT_DISPLAYED = ("not_displayed", "Not displayed")
    ADVANCED_DISPLAY = ("advanced", "Advanced display")
    BASIC_DISPLAY = ("basic", "Basic display")

    def __init__(self, key, display_name):
        self._key = key
        self._disp_name = display_name

    @property
    def key(self):
        """Parameter visibility flag key"""
        return self._key

    @property
    def display_name(self):
        """Parameter visibility display name"""
        return self._disp_name

    @classmethod
    def from_key(cls, key):
        """
        Parameters
        ----------
        key: :obj:`string`
            parameter visibility flag key

        Returns
        -------
        t: :class:`~astrophysix.simdm.experiment.ParameterVisibility`
            Parameter vsibility flag matching the requested key.

        Raises
        ------
        ValueError
            if requested key does not match any parameter visibility.

        Example
        -------
            >>> vis = ParameterVisibility.from_key("advanced")
            >>> vis.display_name
            "Advanced display"
            >>> vis2 = ParameterVisibility.from_key("MY_UNKNOWN_FLAG")
            ValuerError: No ParameterVisibility defined with the key 'MY_UNKNOWN_FLAG'.
        """
        for pvis in cls:
            if pvis.key == key:
                return pvis
        raise ValueError("No ParameterVisibility defined with the key '{k:s}'.".format(k=key))


class ParameterSetting(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Experiment input parameter setting class

    Parameters
    ----------
    input_param: :class:`~astrophysix.simdm.protocol.InputParameter`
        protocol input parameter (mandatory)
    value: :obj:`float` or :obj:`int` or :obj:`string` or :obj:`bool`
        numeric/string/boolean value of the input parameter (mandatory)
    unit: :class:`~astrophysix.units.unit.Unit` or :obj:`string`
        parameter value unit (or unit key :obj:`string`)
    visibility: :class:`~astrophysix.simdm.experiment.ParameterVisibility`
        Parameter setting visibility (for display use only).
        Default :attr:`~astrophysix.simdm.experiment.ParameterVisibility.BASIC_DISPLAY`
    """
    def __init__(self, **kwargs):
        super(ParameterSetting, self).__init__(**kwargs)

        # Protocol associated input parameter
        if "input_param" not in kwargs:
            raise AttributeError("ParameterSetting 'input_param' attribute is not defined (mandatory).")
        inpparam = kwargs["input_param"]
        if not isinstance(inpparam, InputParameter):
            raise AttributeError("Parameter setting 'input_param' attribute is not a valid InputParameter object.")
        self._inpparam = inpparam

        # Input parameters settings
        self._dtype = DataType.REAL
        self._float_value = None
        self._int_value = None
        self._string_value = ""
        self._bool_value = None
        if "value" not in kwargs:
            raise AttributeError("ParameterSetting 'value' attribute is not defined (mandatory).")
        self.value = kwargs["value"]

        self._visibility = ParameterVisibility.BASIC_DISPLAY
        if "visibility" in kwargs:
            self.visibility = kwargs["visibility"]

        self._unit = U.none
        if "unit" in kwargs:
            self.unit = kwargs["unit"]

    def __eq__(self, other):
        """
        ParameterSetting comparison method

        other: :class:`~astrophysix.simdm.experiment.ParameterSetting`
            parameter setting to compare to
        """
        if not super(ParameterSetting, self).__eq__(other):
            return False

        if self._inpparam != other.input_parameter:
            return False

        if self._dtype != other.value_type or self.value != other.value:
            return False

        if self._visibility != other.visibility:
            return False

        return self._unit == other.unit and self._unit.name == other.unit.name

    @property
    def input_parameter(self):
        """Experiment protocol's :class:`~astrophysix.simdm.protocol.InputParameter`. Cannot be edited after parameter
        setting initialisation."""
        return self._inpparam

    @property
    def parameter_key(self):
        """Experiment protocol's :class:`~astrophysix.simdm.protocol.InputParameter` key"""
        return self._inpparam.key

    @property
    def value_type(self):
        """Parameter value type (:class:`~astrophysix.simdm.utils.DataType`)"""
        return self._dtype

    @property
    def value(self):
        """Parameter value.

        Can be set to a :obj:`bool`, :obj:`string`, :obj:`int` or :obj:`float` value. When set,
        :attr:`~astrophysix.simdm.experiment.ParameterSetting.value_type` is also set accordingly.

        Example
        -------
        >>> psetting = ParameterSetting(input_param=inpp, value=0.5)
        >>> type(psetting.value) is float and psetting.value == 0.5 and psetting.value_type == DataType.REAL
        True
        >>> psetting.value = "true"
        >>> type(psetting.value) is bool and psetting.value is True and psetting.value_type == DataType.BOOLEAN
        True
        >>> psetting.value = "false"
        >>> type(psetting.value) is bool and psetting.value is False and psetting.value_type == DataType.BOOLEAN
        True
        >>> psetting.value = "banana"
        >>> type(psetting.value) is str and psetting.value == "banana" and psetting.value_type == DataType.STRING
        True
        >>> psetting.value = 4.256
        >>> type(psetting.value) is float and psetting.value == 4.256 and psetting.value_type == DataType.REAL
        True
        >>> psetting.value = 58.0
        >>> type(psetting.value) is int and psetting.value == 58 and psetting.value_type == DataType.INTEGER
        True
        >>> psetting.value = "3.584e2"
        >>> type(psetting.value) is float and psetting.value == 358.4 and psetting.value_type == DataType.REAL
        True
        >>> psetting.value = "-254"
        >>> type(psetting.value) is int and psetting.value == -254 and psetting.value_type == DataType.INTEGER
        True
        """
        if self._dtype == DataType.REAL:
            return self._float_value
        elif self._dtype == DataType.INTEGER:
            return self._int_value
        elif self._dtype == DataType.BOOLEAN:
            return self._bool_value
        else:  # self._dtype == DataType.STRING:
            return self._string_value

    @value.setter
    def value(self, new_value):
        if isinstance(new_value, bool):  # Boolean value
            self._bool_value = new_value
            self._dtype = DataType.BOOLEAN
            return

        try:
            # Try to cast it as a string
            self._string_value = Stringifiable.cast_string(new_value)
            self._dtype = DataType.STRING

            if self._string_value in ["true", "True", ".true.", "TRUE"]:  # String that can be casted into a boolean (True) value
                self._bool_value = True
                self._dtype = DataType.BOOLEAN
                return
            elif self._string_value in ["false", "False", ".false.", "FALSE"]:  # String that can be casted into a boolean (False) value
                self._bool_value = False
                self._dtype = DataType.BOOLEAN
                return

            # Try to cast it into a numeric value
            try:
                self._cast_numeric_value(self._string_value)
            except ValueError:
                # Not a numeric value, keep it a string value
                pass

        except TypeError:  # Not a valid string object
            try:
                self._cast_numeric_value(new_value)
            except TypeError:  # Not a numeric/string value that can be casted into a real number
                raise AttributeError("ParameterSetting 'value' attribute is not a valid string / integer / float / "
                                     "boolean value")

    def _cast_numeric_value(self, v):
        # Cast it into a float
        fvalue = float(v)
        if fvalue > 2 ** 31 or fvalue < -2 ** 31:
            self._float_value = fvalue
            self._dtype = DataType.REAL
        else:
            ivalue = int(fvalue)  # Convert the float into a int
            if ivalue - fvalue == 0.0:
                self._int_value = ivalue
                self._dtype = DataType.INTEGER
            else:
                self._float_value = fvalue
                self._dtype = DataType.REAL

    @property
    def unit(self):
        """Parameter value unit (:class:`~astrophysix.units.unit.Unit`). Can be edited.

        Example
        -------
        >>> from astrophysix import units as U
        >>> psetting = ParameterSetting(input_param=inpp, value=0.5, unit=U.pc)
        >>> psetting.unit = U.kpc
        >>> psetting.unit = "Mpc"
        """
        return self._unit

    @unit.setter
    def unit(self, new_unit):
        if isinstance(new_unit, U.Unit):
            self._unit = new_unit
        else:
            try:
                s = Stringifiable.cast_string(new_unit, valid_empty=False)
                self._unit = U.Unit.from_name(s)
            except TypeError:  # Not a valid string
                err_msg = "Parameter setting 'unit' property is not a valid (non-empty) string."
                log.error(err_msg)
                raise AttributeError(err_msg)
            except AttributeError as aerr:
                err_msg = "Parameter setting 'unit' property error : {uerr:s}.".format(uerr=str(aerr))
                log.error(err_msg)
                raise AttributeError(err_msg)

    @property
    def visibility(self):
        """Parameter setting visibility flag (:class:`~astrophysix.simdm.experiment.ParameterVisibility`). Can be edited.

        Example
        -------
        >>> psetting = ParameterSetting(input_param=inpp, value=0.5, visibility=ParameterVisibility.ADVANCED_DISPLAY)
        >>> psetting.visibility = ParameterVisibility.BASIC_DISPLAY
        >>> psetting.visibility = "not_displayed"
        """
        return self._visibility

    @visibility.setter
    def visibility(self, new_vis):
        try:
            pvis = Stringifiable.cast_string(new_vis)
            self._visibility = ParameterVisibility.from_key(pvis)
        except ValueError as ve:
            err_msg = "Parameter setting 'visibility' property error : {verr:s}".format(verr=str(ve))
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:  # Not a valid string
            if not isinstance(new_vis, ParameterVisibility):
                err_msg = "ParameterSetting 'visibility' attribute is not a valid ParameterVisibility enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._visibility = new_vis

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["input_param", "value", "unit", "visibility"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a ParameterSetting object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the ParameterSetting into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(ParameterSetting, self)._hsp_write(h5group, **kwargs)

        # Write input parameter UUID
        self._hsp_write_attribute(h5group, ('inpparam_uid', self._inpparam.uid), **kwargs)

        # Write parameter setting value, if defined
        self._hsp_write_attribute(h5group, ('value', self.value), **kwargs)
        self._hsp_write_attribute(h5group, ('value_type', self._dtype.key), **kwargs)

        # Write parameter setting visibility key
        self._hsp_write_attribute(h5group, ('visibility', self._visibility.key), **kwargs)

        # Write parameter setting unit
        self._hsp_write_attribute(h5group, ('unit', self._unit), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a ParameterSetting object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        psetting: ``ParameterSetting``
            Read ParameterSetting instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(ParameterSetting, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read algorithm type
        param_uid = uuid.UUID(cls._hsp_read_attribute(h5group, 'inpparam_uid', "input parameter UUID"))
        protocol = list(dependency_objdict.values())[0]
        found_param = protocol.input_parameters.find_by_uid(param_uid)
        if found_param is None:
            err_msg = "Unknown input parameter UUID '{puid!s}' in the {proto!s}.".format(puid=param_uid, proto=protocol)
            log.error(err_msg)
            raise IOError(err_msg)

        # Read parameter setting value
        v = cls._hsp_read_attribute(h5group, 'value', "parameter setting value")
        vtype = DataType.from_key(cls._hsp_read_attribute(h5group, "value_type", "parameter setting value datatype"))
        if vtype == DataType.BOOLEAN:
            v = bool(v)

        # Read parameter setting visibility
        vis = cls._hsp_read_attribute(h5group, "visibility", "parameter setting visibility")

        # Read parameter setting unit
        u = cls._hsp_read_unit(h5group, "unit")

        # Create parameter setting object
        psetting = cls(uid=uid, input_param=found_param, value=v, visibility=vis, unit=u)

        return psetting

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        # Check that simulation execution time is defined
        if self._dtype == DataType.STRING and len(self._string_value) > 64:
            log.warning("{p!s} Galactica string value is too long (max. 64 characters).".format(p=self))

    def __unicode__(self):
        """
        String representation of the instance
        """
        if self._unit.name != "none":
            us = " {un:s}".format(un=self._unit.name)
        else:
            us = ""
        s = "[{ps_ipnaame:s} = {v!s}{us:s}]".format(ps_ipnaame=self._inpparam.name, v=self.value, us=us)
        s += " parameter setting"
        return s


__all__ = ["ParameterSetting", "ParameterVisibility"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, int
import logging

from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable
from ..datafiles import Datafile
from ..utils import ObjectList, GalacticaValidityCheckMixin
from ..catalogs import Catalog

log = logging.getLogger("astrophysix.simdm")


class GenericResult(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Experiment generic result class

    Parameters
    ----------
    name: :obj:`string`
        result name (mandatory)
    description: :obj:`string`
        result description
    directory_path: :obj:`string`
        result data directory path
    """
    _hsp_version = 2  # with Catalogs
    def __init__(self, **kwargs):
        super(GenericResult, self).__init__(**kwargs)
        self._directory_path = ""
        self._name = ""
        self._description = ""
        self._datafiles = ObjectList(Datafile, "name")
        self._catalogs = ObjectList(Catalog, "name", object_addition_vcheck=(self._can_add_data_proc_service_in_catalog,
                                                                             "processing_services"))
        self._catalogs.add_validity_check_method(self._can_add_catalog)

        # Generic result name
        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        if "description" in kwargs:
            self.description = kwargs["description"]

        if "directory_path" in kwargs:
            self.directory_path = kwargs["directory_path"]

    def __eq__(self, other):
        """
        GenericResult comparison method

        other: :class:`~astrophysix.simdm.results.generic.GenericResult`
            generic result to compare to
        """
        if not super(GenericResult, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._description != other.description:
            return False

        if self._directory_path != other.directory_path:
            return False

        #  Compare datafiles
        if self._datafiles != other.datafiles:
            return False

        # Compare object catalogs
        if self._catalogs != other.catalogs:
            return False

        return True

    @property
    def name(self):
        """Result name. Can be set to a non-empty :obj:`string` value."""
        return self._name

    @name.setter
    def name(self, new_res_name):
        try:
            self._name = Stringifiable.cast_string(new_res_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def description(self):
        """Result description. Can be set to any :obj:`string` value."""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def datafiles(self):
        """Result :class:`~astrophysix.simdm.datafiles.Datafile` list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._datafiles

    @property
    def catalogs(self):
        """Result :class:`~astrophysix.simdm.catalogs.Catalog` list (:class:`~astrophysix.simdm.utils.ObjectList`)

        *New in version 0.5.0*"""
        return self._catalogs

    @property
    def directory_path(self):
        """Result directory.path. Can be set to any :obj:`string` value."""
        return self._directory_path

    @directory_path.setter
    def directory_path(self, new_path):
        try:
            self._directory_path = Stringifiable.cast_string(new_path)
        except TypeError:
            err_msg = "{cname:s} 'directory_path' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_add_catalog(self, new_cat):
        """
        Catalog addition validity check method. Verifies that the catalog added into this GenericResult is not lonked to
        any ata processing service. Otherwise raises an AttributeError.

        Parameters
        ----------
        new_cat: :class:`~astrophysix.simdm.catalogs.catalog.Catalog`
            new Catalog to add into the GenericResult
        """
        if len(new_cat.processing_services) > 0:
            err_msg = "Cannot add a {cat!s} with any data-processing service linked to it into a " \
                      "{cn:s}.".format(cat=new_cat, cn=self.__class__.__name__)
            log.error((err_msg))
            raise AttributeError(err_msg)

    def _can_add_data_proc_service_in_catalog(self, cat_data_proc_service):
        """
        CatalogDataProcessingService addition validity check method. Verifies that the added catalog data processing
        service is not added into a GenericResult's catalog. Otherwise raises an AttributeError.

        Parameters
        ----------
        cat_data_proc_service: :class:`~astrophysix.simdm.services.process.CatalogDataProcessingService`
            new catalog <=> data processing service to add into this Catalog.
        """
        err_msg = "Cannot add a data-processing service into a GenericResult's catalog."
        log.error((err_msg))
        raise AttributeError(err_msg)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "description", "directory_path"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a GenericResult object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the GenericResult into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(GenericResult, self)._hsp_write(h5group, **kwargs)

        # Write result name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write result directory path, if defined
        self._hsp_write_attribute(h5group, ('directory_path', self._directory_path), **kwargs)

        # Write result description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write datafiles, if any defined
        self._hsp_write_object_list(h5group, "DATAFILES", self._datafiles, "datafile_", **kwargs)

        # Write products, if any defined
        self._hsp_write_object_list(h5group, "CATALOGS", self._catalogs, "catalog_", **kwargs)

        self._hsp_write_callback(str(self), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a GenericResult object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        res: ``GenericResult``
            Read GenericResult instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(GenericResult, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read generic result name
        name = cls._hsp_read_attribute(h5group, 'name', "generic result name")

        # Create generic result object
        res = cls(uid=uid, name=name)

        # Read generic result description, if defined
        res_descr = cls._hsp_read_attribute(h5group, 'description', "generic result description",
                                            raise_error_if_not_found=False)
        if res_descr is not None:
            res.description = res_descr

        # Read generic result directory path, if defined
        res_dpath = cls._hsp_read_attribute(h5group, 'directory_path', "generic result directory path",
                                            raise_error_if_not_found=False)
        if res_dpath is not None:
            res.directory_path = res_dpath

        # Build datafile list and add each datafile into generic result
        if "DATAFILES" in h5group:
            for df in Datafile._hsp_read_object_list(h5group, "DATAFILES", "datafile_", "result datafile",
                                                     dependency_objdict=dependency_objdict):
                res.datafiles.add(df)

        # Read catalogs (version >= 2)
        if version >= 2:
            for cat in Catalog._hsp_read_object_list(h5group, "CATALOGS", "catalog_", "catalog",
                                                     dependency_objdict=dependency_objdict):
                res.catalogs.add(cat)

        return res

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check generic result name length
        if len(self._name) > 64:
            log.warning("{r!s} name is too long for Galactica (max. 64 characters).".format(r=self))

        # Perform Galactica validity checks on datafile/catalog list
        self._datafiles.galactica_validity_check(**kwargs)

        # catalog list validity checks
        self._catalogs.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{res_name:s}' generic result".format(res_name=self._name)


__all__ = ["GenericResult"]



# -*- coding: utf-8 -*-
# This software is part of the 'astrophysix' software project.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@author: Damien CHAPON (damien.chapon@cea.fr)
"""
import sys
import argparse
import os
import tempfile
if sys.version_info.major == 2:
    from SocketServer import TCPServer
    from SimpleHTTPServer import SimpleHTTPRequestHandler
else:
    from http.server import SimpleHTTPRequestHandler
    from socketserver import TCPServer

# import webbrowser
import logging

from astrophysix.simdm.datafiles.plot import PlotInfo

D3_WEB_DIR = os.path.abspath(os.path.dirname(__file__))
os.chdir(D3_WEB_DIR)

log = logging.getLogger()
log.setLevel(logging.DEBUG)
_info_formatter = logging.Formatter(fmt="[%(levelname)s] %(message)s")
_console_handler = logging.StreamHandler(stream=sys.stdout)
_console_handler.setLevel(logging.DEBUG)
_console_handler.setFormatter(_info_formatter)
log.addHandler(_console_handler)


class GalD3HttpRequestHandler(SimpleHTTPRequestHandler):
    LOCAL_PLOT_DATA_JSON_FNAME = "plot_data.json"

    def do_GET(self):
        if self.path == '/chart':
            log.info(" -> fetch HTML page (no MathJax.js)")
            self.path = 'plot_no_mathjax.html'
        elif self.path == "/img/loading_icon.png":
            self.path = "loading.png"
        elif self.path == "/img/d3_plot.gif":
            self.path = "d3_plot.gif"
        elif self.path == "/data/plot.json":
            self.path = self.LOCAL_PLOT_DATA_JSON_FNAME

        return SimpleHTTPRequestHandler.do_GET(self)


class GalD3HttpRequestHandlerMathJax(GalD3HttpRequestHandler):
    def do_GET(self):
        if self.path == '/chart':
            log.info(" -> fetch HTML page (with MathJax.js)")
            self.path = 'plot.html'
        elif self.path == "/js/gald3.js":
            log.info(" -> Fetch GalD3.js local JS script")
            self.path = "gald3.js"
        elif self.path == "/static/MathJax_local.js?V=2.7.4":
            log.info(" -> Fetch MathJax local JS config")
            self.path = 'MathJax_local.js?V=2.7.4'

        return GalD3HttpRequestHandler.do_GET(self)


def main():
    """The main routine."""
    log.info("# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #")
    log.info("# ~~~~~~~~~~~~~~~~~~~~ Galactica D3js rendering preview server ~~~~~~~~~~~~~~~~~~~~~~ #")
    log.info("# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #")
    log.info("Press Ctrl+C to terminate the HTTP server.")
    # Do argument parsing here (eg. with argparse) and anything else you want your project to do. Return values are
    # exit codes.

    parser = argparse.ArgumentParser(description="Run GalD3 plot preview HTTP server for the Galactica astrophysical "
                                                 "simulation database")
    parser.add_argument('--debug', dest='debug', action='store_true', help='Run in debug mode')
    parser.add_argument("-n", "--no-mathjax", dest="disable_mathjax", action="store_true", help="Disable MathJax")
    parser.add_argument("-p", "--port", dest="http_port", type=int, default=8076, help="HTTP ser port (default 8076).")
    pargs = parser.parse_args()

    if pargs.disable_mathjax:
        handler = GalD3HttpRequestHandler
    else:
        handler = GalD3HttpRequestHandlerMathJax
    # is_debug = pargs.debug

    tmp_file_path = os.path.join(tempfile.gettempdir(), PlotInfo.PLOT_DATA_JSON_FPATH)
    if not os.path.islink(handler.LOCAL_PLOT_DATA_JSON_FNAME):
        log.info("Linking to temporary JSON data file '{fp:s}'.".format(fp=tmp_file_path))
        os.symlink(tmp_file_path, handler.LOCAL_PLOT_DATA_JSON_FNAME)

    PORT = pargs.http_port
    with TCPServer(("", PORT), handler) as serv:
        log.info("Serving plot preview page at : http://127.0.0.1:{p:d}/chart".format(p=pargs.http_port))

        # Star the server
        # webbrowser.open("http://127.0.0.1:{port:d}/chart".format(port=PORT))
        try:
            serv.serve_forever()
        except KeyboardInterrupt:
            log.info("Server stopped...")
            if os.path.islink(handler.LOCAL_PLOT_DATA_JSON_FNAME):
                log.info("Deleting link to temporary JSON data file '{fp:s}'.".format(fp=tmp_file_path))
                os.remove(handler.LOCAL_PLOT_DATA_JSON_FNAME)
            log.info("# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #")
    return 0


if __name__ == "__main__":
    sys.exit(main())
